# coding: utf-8
#/*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""
This module provides global definitions and functions to manage dark and flat fields
especially for tomo experiments and workflows
"""

__authors__ = ["C. Nemoz", "H.Payno"]
__license__ = "MIT"
__date__ = "06/09/2017"

import functools
import os
import re
import h5py
from glob import glob
import fabio
import numpy
from queue import Queue
from tomwer.utils import docstring
from silx.io.utils import get_data
from silx.io.url import DataUrl
from tomwer.core import settings
from tomwer.core import utils
from tomwer.core.process.baseprocess import SingleProcess, _input_desc, \
    _output_desc
from tomwer.core.signal import Signal
from tomwer.core.utils import getDARK_N
from tomwer.web.client import OWClient
from tomwer.core.log import TomwerLogger
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from . import params as dkrf_reconsparams
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from tomwer.core.scan.scanfactory import ScanFactory
from tomoscan.esrf.utils import get_compacted_dataslices
import tomwer.version
import typing

logger = TomwerLogger(__name__)

try:
    from tomwer.synctools.rsyncmanager import RSyncManager
    has_rsync = False
except ImportError:
    logger.warning('rsyncmanager not available')
    has_rsync = True


class DarkRefs(SingleProcess, Queue):
    """Compute median/mean dark and ref from originals (dark and ref files)
    """

    inputs = [_input_desc(name='data', type=TomwerScanBase, handler='process',
                          doc='scan path'), ]

    outputs = [_output_desc(name='data', type=TomwerScanBase, doc='scan path'), ]

    WHAT_REF = 'refs'
    WHAT_DARK = 'dark'

    VALID_WHAT = (WHAT_REF, WHAT_DARK)
    """Tuple of valid option for What"""

    info_suffix = '.info'

    sigScanReady = Signal(TomwerScanBase)

    sigUpdated = Signal()
    """Emitted when updated when reconstruction parameters are changed"""

    def __init__(self, reconsparams=None, file_ext='.edf'):
        """

        :param str file_ext:
        :param reconsparams: reconstruction parameters
        :type reconsparams: Union[None, ReconsParams, DKRFRP]
        """
        assert type(file_ext) is str
        SingleProcess.__init__(self, )
        Queue.__init__(self)
        self._recons_params = None
        assert reconsparams is None or isinstance(reconsparams, (ReconsParams, dkrf_reconsparams.DKRFRP))

        _recons_params = reconsparams
        if _recons_params is None:
            _recons_params = dkrf_reconsparams.DKRFRP()
        self.set_recons_params(recons_params=_recons_params)

        self._forceSync = False
        if hasattr(self._recons_params, 'sigChanged'):
            self._recons_params.sigChanged.connect(self._updateReconsParam)

        assert isinstance(self._recons_params, dkrf_reconsparams.DKRFRP)
        self._file_ext = file_ext

        self._updateReconsParam()

        self.worker = self._createThread()

    @docstring(SingleProcess.program_name)
    @staticmethod
    def program_name():
        return 'tomwer_dark_refs'

    @docstring(SingleProcess.program_version)
    @staticmethod
    def program_version():
        return tomwer.version.version

    @docstring(SingleProcess.definition)
    @staticmethod
    def definition():
        return 'Compute mean or median dark and refs per each serie'

    def set_recons_params(self, recons_params):
        if isinstance(recons_params, ReconsParams):
            self._recons_params = recons_params.dkrf
        elif isinstance(recons_params, dkrf_reconsparams.DKRFRP):
            self._recons_params = recons_params
        else:
            raise TypeError('recons_params should be an instance of '
                            'ReconsParams or DKRFRP')

    def _createThread(self):
        return DarkRefsWorker(process=self)

    def _updateReconsParam(self):
        self.sigUpdated.emit()

    def disconnect(self):
        if hasattr(self._recons_params, 'sigChanged'):
            self._recons_params.sigChanged.disconnect(self._updateReconsParam)

    def connect(self):
        if hasattr(self._recons_params, 'sigChanged'):
            self._recons_params.sigChanged.connect(self._updateReconsParam)

    def setPatternRecons(self, pattern):
        self._patternReconsFile = pattern

    def _signalScanReady(self, scan):
        assert isinstance(scan, TomwerScanBase)
        self.sigScanReady.emit(scan)
        self.execNext()

    def process(self, scan):
        if scan is None:
            return

        if type(scan) is str:
            assert os.path.exists(scan)
            _scan = ScanFactory.create_scan_object(scan_path=scan)
        elif isinstance(scan, TomwerScanBase):
            _scan = scan
        elif isinstance(scan, dict):
            _scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise TypeError('scan should be an instance of TomoBase or path to '
                            'scan dircetory')
        # copy the current parameters to the scan ftseries_recons_params
        if _scan.ftseries_recons_params is None:
            _scan.ftseries_recons_params = self._instanciateReconsParams()
        assert isinstance(self._recons_params, dkrf_reconsparams.DKRFRP)
        assert isinstance(_scan.ftseries_recons_params, ReconsParams)
        assert self._recons_params is not None

        _scan.ftseries_recons_params.copy(self._recons_params)

        # TODO: now we can only pass the scan as it contains `ftseries_recons_params`
        assert _scan.ftseries_recons_params.dkrf is not None
        Queue.put(self, (_scan, _scan.ftseries_recons_params.dkrf))

        if self.canExecNext():
            self.execNext()

        if self._return_dict:
            return _scan.to_dict()
        else:
            return _scan

    def _instanciateReconsParams(self):
        return ReconsParams(empty=True)

    def execNext(self):
        """Launch the next reconstruction if any
        """
        # try catch is needed because a signal can still be emitted event if the
        # QObject has been destroyed.
        try:
            if super(DarkRefs, self).empty():
                return
        except Exception as error:
            logger.warning(error)
        else:
            scan, params = Queue.get(self)

            self._initWorker(scan=scan, params=params, file_ext=self._file_ext)
            self._launchWorker()
            # TODO: should be in scanready. ok for now since only used with
            # synchronization

            self.register_output(key='data', value=scan.path)

    def _launchWorker(self):
        if hasattr(self.worker, 'isRunning'):
            callback = functools.partial(self._signalScanReady,
                                         self.worker.directory)
            self.worker.finished.connect(callback)
            self.worker.start()
            if self._forceSync is True:
                self.worker.wait()
        else:
            self.worker.process()

    def _initWorker(self, scan, params, file_ext):
        assert isinstance(scan, TomwerScanBase)
        assert type(file_ext) is str
        self.worker.init(scan=scan, params=params, file_ext=file_ext)

    def canExecNext(self):
        """
        Can we launch an ftserie reconstruction.
        Reconstruction can't be runned in parallel

        :return: True if no reconstruction is actually running
        """
        if hasattr(self.worker, 'isRunning'):
            return not self.worker.isRunning()
        else:
            return True

    def setForceSync(self, b):
        self._forceSync = True
        self.worker._forceSync = True

    @staticmethod
    def getRefHSTFiles(directory, prefix, file_ext='.edf'):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
        """
        assert isinstance(directory, str)
        res = []
        if os.path.isdir(directory) is False:
            logger.error(directory + ' is not a directory. Cannot extract '
                                     'RefHST files')
            return res

        for file in os.listdir(directory):
            if file.startswith(prefix) and file.endswith(
                    file_ext):
                res.append(os.path.join(directory, file))
                assert os.path.isfile(res[-1])
        return res

    @staticmethod
    def getDarkHSTFiles(directory, prefix, file_ext='.edf'):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
         """
        res = []
        if os.path.isdir(directory) is False:
            logger.error(directory + ' is not a directory. Cannot extract '
                                     'DarkHST files')
            return res
        for file in os.listdir(directory):
            _prefix = prefix
            if prefix.endswith(file_ext):
                _prefix = prefix.rstrip(file_ext)
            if file.startswith(_prefix) and file.endswith(file_ext):
                _file = file.lstrip(_prefix).rstrip(file_ext)
                if _file == '' or _file.isnumeric() is True:
                    res.append(os.path.join(directory, file))
                    assert os.path.isfile(res[-1])
        return res

    @staticmethod
    def getDarkPatternTooltip():
        return 'define the pattern to find, using the python `re` library.\n' \
               'For example: \n' \
               '   - `.*conti_dark.*` to filter files containing `conti_dark` sentence\n' \
               '   - `darkend[0-9]{3,4}` to filter files named `darkend` followed by three or four digit characters (and having the .edf extension)'

    @staticmethod
    def getRefPatternTooltip():
        return 'define the pattern to find, using the python `re` library.\n' \
               'For example: \n' \
               '   - `.*conti_ref.*` for files containing `conti_dark` sentence\n' \
               '   - `ref*.*[0-9]{3,4}_[0-9]{3,4}` to filter files named `ref` followed by any character and ending by X_Y where X and Y are groups of three or four digit characters.'

    @staticmethod
    def properties_help():
        return """
        - refs: 'None', 'Median', 'Average' \n
        - dark: 'None', 'Median', 'Average' \n
        """

    def setProperties(self, properties):
        # No properties stored for now
        if 'dark' in properties:
            self._recons_params.dark_calc_method = properties['dark']
        if 'refs' in properties:
            self._recons_params.ref_calc_method = properties['refs']

    @staticmethod
    def get_darks_frm_process_file(process_file, entry=None,
                                   as_url: bool = False) -> typing.Union[None, dict]:
        """

        :param str process_file: path to the process file
        :param entry: entry to read in the process file if more than one
        :param bool as_url: if true then an url will be used instead of a
                            numpy.array
        :return: dictionary with index in the sequence as key and numpy array
                 as value (or url if as_url set to True)
        """
        if entry is None:
            with h5py.File(process_file, 'r', swmr=True) as h5f:
                entries = DarkRefs._get_process_nodes(root_node=h5f,
                                                      process=DarkRefs)
                if len(entries) == 0:
                    logger.info(
                        'unable to find a DarkRef process in %s' % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError('several entry found, entry should be '
                                     'specify')
                else:
                    entry = list(entries.keys())[0]
                    logger.info('take %s as default entry' % entry)

        with h5py.File(process_file, 'r', swmr=True) as h5f:
            if entry not in h5f.keys():
                logger.warning('no dark found for {}'.format(entry))
                return {}
            dark_nodes = DarkRefs._get_process_nodes(root_node=h5f[entry],
                                                     process=DarkRefs)
            index_to_path = {}
            for key, value in dark_nodes.items():
                index_to_path[key] = key

            if len(dark_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(list(dark_nodes.keys()))[-1]
            last_process_dark = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                logger.debug('several processing found for dark-ref,'
                               'take the last one: %s' % last_process_dark)

            res = {}
            if 'results' in h5f[last_process_dark].keys():
                results_node = h5f[last_process_dark]['results']
                if 'darks' in results_node.keys():
                    darks = results_node['darks']
                    for index in darks:
                        if as_url is True:
                            res[int(index)] = DataUrl(file_path=process_file,
                                                      data_path='/'.join((darks.name, index)),
                                                      scheme='silx')
                        else:
                            res[int(index)] = darks[index][()]
            return res

    @staticmethod
    def get_flats_frm_process_file(process_file, entry=None,
                                   as_url: bool = False) -> typing.Union[None, dict]:
        """

        :param process_file:
        :param entry: entry to read in the process file if more than one
        :param bool as_url: if true then an url will be used instead of a
                            numpy.array
        :return:
        """
        if entry is None:
            with h5py.File(process_file, 'r', swmr=True) as h5f:
                entries = DarkRefs._get_process_nodes(root_node=h5f,
                                                      process=DarkRefs)
                if len(entries) == 0:
                    logger.info(
                        'unable to find a DarkRef process in %s' % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError('several entry found, entry should be '
                                     'specify')
                else:
                    entry = list(entries.keys())[0]
                    logger.info('take %s as default entry' % entry)

        with h5py.File(process_file, 'r', swmr=True) as h5f:
            if entry not in h5f.keys():
                logger.warning('no flats found for {}'.format(entry))
                return {}
            dkref_nodes = DarkRefs._get_process_nodes(root_node=h5f[entry],
                                                      process=DarkRefs)
            index_to_path = {}
            for key, value in dkref_nodes.items():
                index_to_path[key] = key

            if len(dkref_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(dkref_nodes.keys())[-1]
            last_process_dkrf = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                logger.debug('several processing found for dark-ref,'
                               'take the last one: %s' % last_process_dkrf)

            res = {}
            if 'results' in h5f[last_process_dkrf].keys():
                results_node = h5f[last_process_dkrf]['results']
                if 'flats' in results_node.keys():
                    flats = results_node['flats']
                    for index in flats:
                        if as_url is True:
                            res[int(index)] = DataUrl(file_path=process_file,
                                                      data_path='/'.join((flats.name, index)),
                                                      scheme='silx')
                        else:
                            res[int(index)] = flats[index][()]
            return res


class DarkRefsWorker(OWClient):
    """Worker of the Dark refs"""

    TOMO_N = "TOMO_N"

    def __init__(self, process):
        OWClient.__init__(self, logger)
        self.recons_params = None
        self._forceSync = False
        self.scan = None
        self.directory = None
        self._process = process

    def init(self, scan, params, file_ext):
        assert isinstance(scan, TomwerScanBase)
        assert params is not None
        assert isinstance(params, dkrf_reconsparams.DKRFRP)
        assert type(file_ext) is str
        self.scan = scan
        self.directory = self.scan.path
        self.recons_params = params
        self._file_ext = file_ext

    def run(self):
        self.process()

    def process(self):
        if settings.isOnLbsram() and utils.isLowOnMemory(
                settings.get_lbsram_path()) is True:
            mess = 'low memory, do compute dark and flat field mean/median ' \
                   'for %s' % self.scan.path
            logger.processSkipped(mess)
            return

        if not(self.scan and
               os.path.exists(self.scan.path) and
               os.path.isdir(self.scan.path)):
            logger.warning("folder %s is not existing" % self.scan.path)
            return
        whats = (DarkRefs.WHAT_REF, DarkRefs.WHAT_DARK)
        modes = (self.recons_params.ref_calc_method, self.recons_params.dark_calc_method)
        for what, mode in zip(whats, modes):
            self._originalsDark = []
            self._originalsRef = []
            self.compute(scan=self.scan, what=what, mode=mode)
        if self.scan.process_file is not None:
            results = {}
            interpretations = {}
            if self.scan.normed_darks is not None:
                # cast darks and flats keys from int (index) to str
                o_darks = self.scan.normed_darks
                f_darks = {}
                for index, data in o_darks.items():
                    f_darks[str(index)] = data
                    interpretations['/'.join(('darks', str(index)))] = 'image'
                results['darks'] = f_darks
            if self.scan.normed_flats is not None:
                results['flats'] = self.scan.normed_flats
                o_flats = self.scan.normed_flats
                f_flats = {}
                for index, data in o_flats.items():
                    f_flats[str(index)] = data
                    interpretations['/'.join(('flats', str(index)))] = 'image'
                results['flats'] = f_flats

            if len(results) > 0:
                process = self._process
                with self.scan.acquire_process_file_lock():
                    entry = 'entry'
                    if hasattr(self.scan, 'entry'):
                        entry = self.scan.entry
                    process.register_process(process_file=self.scan.process_file,
                                             entry=entry,
                                             configuration=self.recons_params.to_dict(),
                                             results=results,
                                             interpretations=interpretations,
                                             process_index=self.scan.pop_process_index(),
                                             overwrite=True)

    def compute(self, scan, what, mode):
        if isinstance(scan, EDFTomoScan):
            self.compute_edf(scan, what=what, mode=mode)
        elif isinstance(scan, HDF5TomoScan):
            self.compute_hdf5(scan, what=what, mode=mode)
        else:
            raise ValueError('scan type is not recognized ofr %s' % scan)

    def compute_hdf5(self, scan, what, mode):
        """Compute the requested what in the given mode for `directory`

        :param str directory: path of the scan
        :param what: what to compute (ref or dark)
        :param mode: how to compute it (median or average...)
        """
        def get_series(scan, what: str) -> list:
            """return a list of dictionary. on the dictionary we have indexes
            as key and url as value"""
            if what == 'dark':
                raw_what = scan.darks
            else:
                raw_what = scan.flats
            if len(raw_what) == 0:
                return []
            else:
                series = []
                indexes = sorted(raw_what.keys())
                # a serie is defined by contiguous indexes
                current_serie = {indexes[0]: raw_what[indexes[0]]}
                current_index = indexes[0]
                for index in indexes[1:]:
                    if index == current_index + 1:
                        current_index = index
                    else:
                        series.append(current_serie)
                        current_serie = {}
                        current_index = index
                    current_serie[index] = raw_what[index]
                if len(current_serie) > 0:
                    series.append(current_serie)
                return series

        if mode is dkrf_reconsparams.Method.median:
            method_ = numpy.median
        elif mode is dkrf_reconsparams.Method.average:
            method_ = numpy.mean
        else:
            return
        raw_series = get_series(scan, what)
        if len(raw_series) == 0:
            logger.info('No %s found for %s' % (what, scan))
            return

        def load_data_serie(urls):
            
            cpt_slices = get_compacted_dataslices(urls)
            url_set = {}
            for url in cpt_slices.values():
                path = url.file_path(), url.data_path(), str(url.data_slice())
                url_set[path] = url
			
            n_elmts = 0
            for url in url_set.values():
                my_slice = url.data_slice()
                n_elmts += my_slice.stop - my_slice.start
            	
            data = None
            start_z = 0
            for url in url_set.values():
                my_slice = url.data_slice()
                my_slice = slice(my_slice.start, my_slice.stop, 1)
                new_url = DataUrl(file_path=url.file_path(),
                                  data_path=url.data_path(),
                                  data_slice=my_slice,
                                  scheme='silx')
                loaded_data = get_data(new_url)
                
                # init data if dim is not know
                if data is None:
                    data = numpy.empty(shape=(n_elmts,
                                              scan.dim_2 or loaded_data.shape[-2],
                                              scan.dim_1 or loaded_data.shape[-1]))
                if loaded_data.ndim == 2:
                    data[start_z, :, :] = loaded_data
                    start_z += 1
                elif loaded_data.ndim == 3:
                    delta_z = my_slice.stop - my_slice.start
                    data[start_z:delta_z, :, :] = loaded_data
                    start_z += delta_z
                else:
                    raise ValueError('Dark and ref raw data should be 2D or 3D')
            
            return data

        res = {}
        # res: index: sequence when the serie was taken

        for serie_ in raw_series:
            serie_index = min(serie_)
            serie_data = load_data_serie(serie_)
            res[serie_index] = method_(serie_data, axis=0)
        if what == 'dark':
            scan.set_normed_darks(res)
        else:
            scan.set_normed_flats(res)

    def compute_edf(self, scan, what, mode):
        """Compute the requested what in the given mode for `directory`

        :param str directory: path of the scan
        :param what: what to compute (ref or dark)
        :param mode: how to compute it (median or average...)
        """
        directory = scan.path
        assert type(directory) is str

        def removeFiles():
            """Remove orignals files fitting the what (dark or ref)"""
            if what is DarkRefs.WHAT_DARK:
                # In the case originals has already been found for the median
                # calculation
                if len(self._originalsDark) > 0:
                    files = self._originalsDark
                else:
                    files = getOriginals(DarkRefs.WHAT_DARK)
            elif what is DarkRefs.WHAT_REF:
                if len(self._originalsRef) > 0:
                    files = self._originalsRef
                else:
                    files = getOriginals(DarkRefs.WHAT_REF)
            else:
                logger.error(
                    'the requested what (%s) is not recognized. '
                    'Can\'t remove corresponding file' % what)
                return

            _files = set(files)
            if len(files) > 0:
                if has_rsync:
                    logger.info('ask RSyncManager for removal of %s files in %s' % (what, directory))
                    # for lbsram take into account sync from data watcher
                    if directory.startswith(settings.get_lbsram_path()):
                        for f in files:
                            _files.add(f.replace(settings.get_lbsram_path(),
                                                 settings.get_dest_path(),
                                                 1))
                    RSyncManager().removesync_files(dir=directory,
                                                   files=_files,
                                                   block=self._forceSync)
                else:
                    for _file in _files:
                        try:
                            os.remove(_file)
                        except Exception as e:
                            logger.error(e)

        def getOriginals(what):
            if what is DarkRefs.WHAT_REF:
                try:
                    pattern = re.compile(self.recons_params.ref_pattern)
                except:
                    pattern = None
                    logger.error('Fail to compute regular expresion for %s' % self.recons_params.ref_pattern)
            elif what is DarkRefs.WHAT_DARK:
                re.compile(self.recons_params.dark_pattern)
                try:
                    pattern = re.compile(self.recons_params.dark_pattern)
                except:
                    pattern = None
                    logger.error('Fail to compute regular expresion for %s' % self.recons_params.dark_pattern)

            filelist_fullname = []
            if pattern is None:
                return filelist_fullname
            for file in os.listdir(directory):
                if pattern.match(file) and file.endswith(self._file_ext):
                    if (file.startswith(self.recons_params.ref_prefix) or
                            file.startswith(self.recons_params.dark_prefix)) is False:
                        filelist_fullname.append(os.path.join(directory, file))
            return sorted(filelist_fullname)

        def setup():
            """setup parameter to process the requested what

            :return: True if there is a process to be run, else false
            """
            def getNDigits(_file):
                file_without_scanID = _file.replace(os.path.basename(directory), '', 1)
                return len(re.findall(r'\d+', file_without_scanID))

            def dealWithPCOTomo():
                filesPerSerie = {}
                if self.nfiles % self.nacq == 0:
                    assert self.nacq < self.nfiles
                    self.nseries = self.nfiles // self.nacq
                    self.series = self.fileNameList
                else:
                    logger.warning('Fail to deduce series')
                    return None, None

                linear = getNDigits(self.fileNameList[0]) < 2
                if linear is False:
                    # which digit pattern contains the file number?
                    lastone = True
                    penulti = True
                    for first_files in range(self.nseries - 1):
                        digivec_1 = re.findall(r'\d+', self.fileNameList[
                            first_files])
                        digivec_2 = re.findall(r'\d+', self.fileNameList[
                            first_files + 1])
                        if lastone:
                            lastone = (
                                (int(digivec_2[-1]) - int(digivec_1[-1])) == 0)
                        if penulti:
                            penulti = (
                                (int(digivec_2[-2]) - int(digivec_1[-2])) == 0)

                    linear = not penulti

                if linear is False:
                    digivec_1 = re.findall(r'\d+', self.fileNameList[
                        self.nseries - 1])
                    digivec_2 = re.findall(r'\d+',
                                           self.fileNameList[self.nseries])
                    # confirm there is 1 increment after self.nseries in the uperlast last digit patern
                    if ((int(digivec_2[-2]) - int(digivec_1[-2])) != 1):
                        linear = True

                # series are simple sublists in main filelist
                # self.series = []
                if linear is True:
                    is_there_digits = len(
                        re.findall(r'\d+', self.fileNameList[0])) > 0
                    if is_there_digits:
                        serievec = set(
                            [re.findall(r'\d+', self.fileNameList[0])[-1]])
                    else:
                        serievec = set(['0000'])
                    for i in range(self.nseries):
                        if is_there_digits:
                            serie = re.findall(r'\d+', self.fileNameList[i * self.nacq])[-1]
                            serievec.add(serie)
                            filesPerSerie[serie] = self.fileNameList[i * self.nacq:(i+1)*self.nacq]
                        else:
                            serievec.add('%04d' % i)
                # in the sorted filelist, the serie is incremented, then the acquisition number:
                else:
                    self.series = self.fileNameList[0::self.nseries]
                    serievec = set(
                        [re.findall(r'\d+', self.fileNameList[0])[-1]])
                    for serie in serievec:
                        # serie = re.findall(r'\d+', self.fileNameList[i])[-1]
                        # serievec.add(serie)
                        filesPerSerie[serie] = self.fileNameList[0::self.nseries]
                serievec = list(sorted(serievec))

                if len(serievec) > 2:
                    logger.error('DarkRefs do not deal with multiple scan.'
                                 ' (scan %s)' % directory)
                    return None, None
                assert len(serievec) <= 2
                if len(serievec) > 1:
                    key = serievec[-1]
                    tomoN = self.getInfo(self.TOMO_N)
                    if tomoN is None:
                        logger.error("Fail to found information %s. Can't "
                                     "rename %s" % (self.TOMO_N, key))
                    del serievec[-1]
                    serievec.append(str(tomoN).zfill(4))
                    filesPerSerie[serievec[-1]] = filesPerSerie[key]
                    del filesPerSerie[key]
                    assert len(serievec) == 2
                    assert len(filesPerSerie) == 2

                return serievec, filesPerSerie

            # start setup function
            if mode == dkrf_reconsparams.Method.none:
                return False
            if what == 'dark':
                self.out_prefix = self.recons_params.dark_prefix
                self.info_nacq = 'DARK_N'
            else:
                self.out_prefix = self.recons_params.ref_prefix
                self.info_nacq = 'REF_N'

            # init
            self.nacq = 0
            """Number of acquisition runned"""
            self.files = 0
            """Ref or dark files"""
            self.nframes = 1
            """Number of frame per ref/dark file"""
            self.serievec = ['0000']
            """List of series discover"""
            self.filesPerSerie = {}
            """Dict with key the serie id and values list of files to compute
            for median or mean"""
            self.infofile = ''
            """info file of the acquisition"""

            # sample/prefix and info file
            self.prefix = os.path.basename(directory)
            extensionToTry = (DarkRefs.info_suffix, '0000' + DarkRefs.info_suffix)
            for extension in extensionToTry:
                infoFile = os.path.join(directory, self.prefix + extension)
                if os.path.exists(infoFile):
                    self.infofile = infoFile
                    break

            if self.infofile == '':
                logger.debug('fail to found .info file for %s' % directory)

            """
            Set filelist
            """
            # do the job only if not already done and overwrite not asked
            self.out_files = sorted(
                glob(directory + os.sep + '*.' + self._file_ext))

            self.filelist_fullname = getOriginals(what)
            self.fileNameList = []
            [self.fileNameList.append(os.path.basename(_file)) for _file in self.filelist_fullname]
            self.fileNameList = sorted(self.fileNameList)
            self.nfiles = len(self.filelist_fullname)
            # if nothing to process
            if self.nfiles == 0:
                logger.info('no %s for %s, because no file to compute found' % (what, directory))
                return False

            self.fid = fabio.open(self.filelist_fullname[0])
            self.nframes = self.fid.nframes
            self.nacq = 0
            # get the info of number of acquisitions
            if self.infofile != '':
                self.nacq = self.getInfo(self.info_nacq)

            if self.nacq == 0:
                self.nacq = self.nfiles

            self.nseries = 1
            if self.nacq > self.nfiles:
                # get ready for accumulation and/or file multiimage?
                self.nseries = self.nfiles
            if self.nacq < self.nfiles and getNDigits(self.fileNameList[0]) < 2:
                self.nFilePerSerie = self.nseries
                self.serievec, self.filesPerSerie = dealWithPCOTomo()
            else:
                self.series = self.fileNameList
                self.serievec = _getSeriesValue(self.fileNameList)
                self.filesPerSerie, self.nFilePerSerie = groupFilesPerSerie(self.filelist_fullname, self.serievec)

            if self.filesPerSerie is not None:
                for serie in self.filesPerSerie:
                    for _file in self.filesPerSerie[serie]:
                        if what == 'dark':
                            self._originalsDark.append(os.path.join(self.directory, _file))
                        elif what == 'ref':
                            self._originalsRef.append(os.path.join(self.directory, _file))

            return self.serievec is not None and self.filesPerSerie is not None

        def _getSeriesValue(fileNames):
            assert (len(fileNames) > 0)
            is_there_digits = len(re.findall(r'\d+', fileNames[0])) > 0
            series = set()
            i = 0
            for fileName in fileNames:
                if is_there_digits:
                    name = fileName.rstrip(self._file_ext)
                    file_index = name.split('_')[-1]
                    rm_not_numeric = re.compile(r'[^\d.]+')
                    file_index = rm_not_numeric.sub('', file_index)
                    series.add(file_index)
                else:
                    series.add('%04d' % i)
                    i += 1
            return list(series)

        def groupFilesPerSerie(files, series):
            def findFileEndingWithSerie(poolFiles, serie):
                res = []
                for _file in poolFiles:
                    _f = _file.rstrip('.edf')
                    if _f.endswith(serie):
                        res.append(_file)
                return res

            def checkSeriesFilesLength(serieFiles):
                length = -1
                for serie in serieFiles:
                    if length == -1:
                        length = len(serieFiles[serie])
                    elif len(serieFiles[serie]) != length:
                        logger.error('Series with inconsistant number of ref files')

            assert len(series) > 0
            if len(series) == 1:
                return {series[0]: files}, len(files)
            assert len(files) > 0

            serieFiles = {}
            unattributedFiles = files.copy()
            for serie in series:
                serieFiles[serie] = findFileEndingWithSerie(unattributedFiles, serie)
                [unattributedFiles.remove(_f) for _f in serieFiles[serie]]

            if len(unattributedFiles) > 0:
                logger.error('Failed to associate %s to any serie' % unattributedFiles)
                return {}, 0

            checkSeriesFilesLength(serieFiles)

            return serieFiles, len(serieFiles[list(serieFiles.keys())[0]])

        def process():
            """process calculation of the what"""
            if mode is dkrf_reconsparams.Method.none:
                return
            shape = fabio.open(self.filelist_fullname[0]).shape

            for i in range(len(self.serievec)):
                largeMat = numpy.zeros((self.nframes * self.nFilePerSerie,
                                        shape[0],
                                        shape[1]))

                if what == 'dark' and len(self.serievec) == 1:
                    fileName = self.out_prefix
                    if fileName.endswith(self._file_ext) is False:
                        fileName = fileName + self._file_ext
                else:
                    fileName = self.out_prefix.rstrip(self._file_ext) \
                               + self.serievec[i] + self._file_ext
                fileName = os.path.join(directory, fileName)
                if os.path.isfile(fileName):
                    if ((what == 'refs' and self.recons_params.overwrite_ref is False) or
                            (what == 'dark' and self.recons_params.overwrite_dark is False)):
                        logger.info(
                            'skip creation of %s, already existing' % fileName)
                        continue

                if self.nFilePerSerie == 1:
                    fSerieName = os.path.join(directory, self.series[i])
                    header = {'method': mode.name + ' on 1 image'}
                    header['SRCUR'] = utils.getClosestSRCurrent(
                        scan_dir=directory,
                        refFile=fSerieName)
                    if self.nframes == 1:
                        largeMat[0] = fabio.open(fSerieName).data
                    else:
                        handler = fabio.open(fSerieName)
                        dShape = (self.nframes, handler.dim2, handler.dim1)
                        largeMat = numpy.zeros(dShape)
                        for iFrame in range(self.nframes):
                            largeMat[iFrame] = handler.getframe(iFrame).data
                else:
                    header = {'method': mode.name + ' on %d images' % self.nFilePerSerie}
                    header['SRCUR'] = utils.getClosestSRCurrent(
                        scan_dir=directory,
                        refFile=self.series[i][0])
                    for j, fName in zip(range(self.nFilePerSerie), self.filesPerSerie[self.serievec[i]]):
                        file_BigMat = fabio.open(fName)
                        if self.nframes > 1:
                            for fr in range(self.nframes):
                                jfr = fr + j * self.nframes
                                largeMat[jfr] = file_BigMat.getframe(
                                    fr).getData()
                        else:
                            largeMat[j] = file_BigMat.data

                if mode == dkrf_reconsparams.Method.median:
                    data = numpy.median(largeMat, axis=0)
                elif mode == dkrf_reconsparams.Method.average:
                    data = numpy.mean(largeMat, axis=0)
                else:
                    assert mode != dkrf_reconsparams.Method.none
                    logger.error(
                        'Unrecognized calculation type request %s' % mode)
                    return
                self.nacq = getDARK_N(directory) or 1
                if what == 'dark' and self.nacq > 1:  # and self.nframes == 1:
                    data = data / self.nacq
                    # add one to add to avoid division by zero
                    # data = data + 1
                file_desc = fabio.edfimage.EdfImage(data=data,
                                                    header=header)
                i += 1
                _ttype = numpy.uint16 if what == 'dark' else numpy.int32
                file_desc.write(fileName, force_type=_ttype)

        if directory is None:
            return
        if setup():
            logger.info('start proccess darks and flat fields for %s' % self.scan.path)
            process()
            logger.info('end proccess darks and flat fields')

        self._store_result(what=what, scan=scan)
        if ((what == 'dark' and self.recons_params.remove_dark is True) or
                (what == 'refs' and self.recons_params.remove_ref is True)):
            removeFiles()

    def _store_result(self, what, scan):
        try:
            if what == 'dark':
                dark = DarkRefs.getDarkHSTFiles(directory=scan.path,
                                                prefix=self.recons_params.dark_prefix)
                dark_1 = fabio.open(dark[0]).data
                scan.set_normed_darks({0: dark_1})
            else:
                refs = DarkRefs.getRefHSTFiles(directory=scan.path,
                                               prefix=self.recons_params.ref_prefix)
                refs_dict = {}
                if len(refs) == 1:
                    ref_1 = fabio.open(refs[0]).data
                    refs_dict[0] = ref_1
                elif len(refs) >= 2:
                    ref_1 = fabio.open(refs[0]).data
                    refs_dict[0] = ref_1
                    ref_f = fabio.open(refs[-1]).data
                    refs_dict[scan.tomo_n] = ref_f
                scan.set_normed_flats(refs_dict)
        except Exception as e:
            logger.info(e)

    def getInfo(self, what):
        with open(self.infofile) as file:
            infod = file.readlines()
            for line in infod:
                if what in line:
                    return int(line.split('=')[1])
        # not found:
        return 0

    def getDarkFiles(self, directory):
        """

        :return: the list of existing darks files in the directory according to
                 the file pattern.
         """
        patternDark = re.compile(self.recons_params.dark_pattern)

        res = []
        for file in os.listdir(directory):
            if patternDark.match(file) is not None and file.endswith(
                    self._file_ext):
                res.append(os.path.join(directory, file))
        return res

    def getRefFiles(self, directory):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
         """
        patternRef = re.compile(self.recons_params.ref_pattern)

        res = []
        for file in os.listdir(directory):
            if patternRef.match(file) and file.endswith(self._file_ext):
                res.append(os.path.join(directory, file))
        return res



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from ..utils import WidgetLongProcessing
from silx.gui import qt
from Orange.widgets import widget, gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Input, Output
from orangecontrib.tomwer.orange.settings import CallbackSettingsHandler
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs, \
    logger as DRLogger
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.reconstruction.darkref.darkrefcopywidget import DarkRefAndCopyWidget
from tomwer.synctools.ftseries import QReconsParams
from tomwer.web.client import OWClient

logger = logging.getLogger(__name__)


class DarkRefAndCopyOW(widget.OWWidget, OWClient, WidgetLongProcessing):
    """
        A simple widget managing the copy of an incoming folder to an other one

        :param parent: the parent widget
        """
    # note of this widget should be the one registred on the documentation
    name = "dark and flat field construction"
    id = "orange.widgets.tomwer.darkrefs"
    description = "This widget will generate dark refs for a received scan "
    icon = "icons/darkref.svg"
    priority = 25
    category = "esrfWidgets"
    keywords = ["tomography", "dark", "darks", "ref", "refs"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    settingsHandler = CallbackSettingsHandler()

    _rpSetting = Setting(dict())
    """Setting to load and save DarkRefAndCopyWidget settings"""

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    assert len(DarkRefs.inputs) == 1
    class Inputs:
        data_in = Input(name=DarkRefs.inputs[0].name,
                        type=DarkRefs.inputs[0].type,
                        doc=DarkRefs.inputs[0].doc)


    assert len(DarkRefs.outputs) == 1
    class Outputs:
        data_out = Output(name=DarkRefs.outputs[0].name,
                          type=DarkRefs.outputs[0].type,
                          doc=DarkRefs.outputs[0].doc)

    def __init__(self, parent=None, _connect_handler=True, reconsparams=None):
        """

        :param bool _connect_handler: True if we want to store the modifications
                                      on the setting. Need for unit test since
                                      keep alive qt widgets.
        :param QReconsParams reconsparams: reconstruction parameters
        """
        widget.OWWidget.__init__(self, parent)
        OWClient.__init__(self, (logger, DRLogger))
        reconsparams = reconsparams or QReconsParams()
        if self._rpSetting != dict():
            try:
                reconsparams.dkrf.load_from_dict(self._rpSetting)
            except:
                logger.warning('fail to load reconstruction settings')

        self.widget = DarkRefAndCopyWidget(parent=self,
                                           reconsparams=reconsparams)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.addWidget(self.widget)
        self.setForceSync = self.widget.setForceSync

        # expose API
        self.hasRefStored = self.widget.hasFlatStored
        self.setModeAuto = self.widget.set_mode_auto
        self.setRefsFromScan = self.widget.setRefsFromScan
        self.setCopyActive = self.widget.setCopyActive
        self.hasDarkStored = self.widget.hasDarkStored
        self.hasFlatStored = self.widget.hasFlatStored

        # connect DarkRef with the widget
        self.widget.sigProcessingStart.connect(self._startProcessing)
        self.widget.sigProcessingEnd.connect(self._endProcessing)
        self.widget.sigScanReady.connect(self.signalReady)
        if _connect_handler:
            self.settingsHandler.addCallback(self._updateSettingsVals)

    @Inputs.data_in
    def process(self, scanID):
        if scanID is None:
            return
        assert isinstance(scanID, TomwerScanBase)
        return self.widget.process(scanID)

    def signalReady(self, scanID):
        assert isinstance(scanID, TomwerScanBase)
        self.Outputs.data_out.send(scanID)
        self.sigScanReady.emit(scanID)

    def _updateSettingsVals(self):
        self._rpSetting = self.widget.recons_params.to_dict()

    @property
    def recons_params(self):
        return self.widget.recons_params

    def close(self):
        logger.info('close Dark refs')
        self.widget.close()
        super(DarkRefAndCopyOW, self).close()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/02/2020"


from silx.gui import qt
from tomwer.gui.reconstruction.nabu.nabuconfig.preprocessing import _NabuPreProcessingConfig
from tomwer.gui.reconstruction.nabu.nabuconfig.phase import _NabuPhaseConfig
from tomwer.gui.reconstruction.nabu.nabuconfig.reconstruction import _NabuReconstructionConfig
from tomwer.gui.reconstruction.nabu.nabuconfig.output import _NabuOutputConfig
from tomwer.core.process.reconstruction.nabu import _NabuStages


class NabuConfiguration(qt.QWidget):
    """
    Top level widget for defining the nabu configuration
    """

    sigConfChanged = qt.Signal(str, str)
    """Signal emitted when the configuration change. Parameters are
    (stage, index option modified)
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent)

        self.setLayout(qt.QVBoxLayout())

        # pre processing options
        self._preProcessingGB = qt.QGroupBox('pre processing', self)
        self._preProcessingGB.setLayout(qt.QVBoxLayout())

        self._preProcessingWidget = _NabuPreProcessingConfig(parent=self)
        self._preProcessingGB.layout().addWidget(self._preProcessingWidget)
        self.layout().addWidget(self._preProcessingGB)

        # phase options
        self._phaseGB = qt.QGroupBox('phase', self)
        self._phaseGB.setLayout(qt.QVBoxLayout())

        self._phaseWidget = _NabuPhaseConfig(parent=self)
        self._phaseGB.layout().addWidget(self._phaseWidget)
        self._phaseGB.setCheckable(True)
        self.layout().addWidget(self._phaseGB)

        # reconstruction opts
        self._reconstructionGB = qt.QGroupBox('reconstruction', self)
        self._reconstructionGB.setLayout(qt.QVBoxLayout())

        self._reconstructionWidget = _NabuReconstructionConfig(parent=self)
        self._reconstructionGB.layout().addWidget(self._reconstructionWidget)
        self.layout().addWidget(self._reconstructionGB)

        # output information
        self._outputGB = qt.QGroupBox('output', self)
        self._outputGB.setLayout(qt.QVBoxLayout())

        self._outputWidget = _NabuOutputConfig(parent=self)
        self._outputGB.layout().addWidget(self._outputWidget)
        self.layout().addWidget(self._outputGB)

        # set up
        self._phaseGB.setChecked(True)

        # connect signal / slot
        self._preProcessingWidget.sigConfChanged.connect(
            self._signalConfChangedPreProcessing)
        self._phaseWidget.sigConfChanged.connect(
            self._signalConfChangedPhase)
        self._reconstructionWidget.sigConfChanged.connect(
            self._signalConfChangedReconstruction)

        # expose API
        self.getSlices = self._reconstructionWidget.getSlices
        self.setOutputDir = self._outputWidget.setOutputDir

    def _signalConfChangedPhase(self, param):
        self.sigConfChanged.emit(self._phaseWidget.getStage().value, param)

    def _signalConfChangedPreProcessing(self, param):
        self.sigConfChanged.emit(self._preProcessingWidget.getStage().value, param)

    def _signalConfChangedReconstruction(self, param):
        self.sigConfChanged.emit(self._reconstructionWidget.getStage().value, param)

    def getConfiguration(self):
        config = {
            'preproc': self._preProcessingWidget.getConfiguration(),
            'reconstruction': self._reconstructionWidget.getConfiguration(),
            'tomwer_slices': self._reconstructionWidget.getSlices(),
            'output': self._outputWidget.getConfiguration(),
            'phase': self._phaseWidget.getConfiguration()
        }
        if not self._phaseGB.isChecked():
            config['phase']['method'] = ''
        return config

    def setConfiguration(self, config):
        if 'preproc' in config:
            self._preProcessingWidget.setConfiguration(config['preproc'])
        if 'phase' in config:
            self._phaseWidget.setConfiguration(config['phase'])
            self._phaseGB.setChecked(True)
        else:
            self._phaseGB.setChecked(False)
        if 'reconstruction' in config:
            self._reconstructionWidget.setConfiguration(config['reconstruction'])
        if 'tomwer_slices' in config:
            self._reconstructionWidget.setSlices(config['tomwer_slices'])

    def applyFilter(self, stage, option):
        if stage is None:
            for widget in (self._preProcessingGB, self._reconstructionGB,
                           self._outputGB, self._phaseGB):
                widget.setVisible(True)
        else:
            stage = _NabuStages.from_value(stage)
            self._preProcessingGB.setVisible(stage is _NabuStages.PRE)
            self._reconstructionGB.setVisible(stage is _NabuStages.PROC)
            self._outputGB.setVisible(stage is _NabuStages.POST)
            self._phaseGB.setVisible(stage is _NabuStages.PHASE)

    def setConfigurationLevel(self, level):
        for widget in (self._preProcessingWidget, self._reconstructionWidget,
                       self._outputWidget, self._phaseWidget):
            widget.setConfigurationLevel(level)


class NabuConfigurationTab(qt.QTabWidget):
    """
    Top level widget for defining the nabu configuration.
    Same as NabuConfiguration but inside a tab
    """

    sigConfChanged = qt.Signal(str, str)
    """Signal emitted when the configuration change. Parameters are
    (stage, index option modified)
    """

    def __init__(self, parent):
        qt.QTabWidget.__init__(self, parent=parent)

        # pre processing options
        self._preProcessingWidget = _NabuPreProcessingConfig(parent=self)
        self.addTab(self._preProcessingWidget, 'pre processing')

        # phase options
        self._phaseGB = qt.QGroupBox('apply phase', self)
        self._phaseGB.setLayout(qt.QVBoxLayout())

        self._phaseWidget = _NabuPhaseConfig(parent=self)
        self._phaseGB.layout().addWidget(self._phaseWidget)
        self._phaseGB.setCheckable(True)
        self.addTab(self._phaseGB, 'phase')

        # reconstruction opts
        self._reconstructionWidget = _NabuReconstructionConfig(parent=self)
        self.addTab(self._reconstructionWidget, 'reconstruction')

        # output information
        self._outputWidget = _NabuOutputConfig(parent=self)
        self.addTab(self._outputWidget, 'output')

        # set up
        self._phaseGB.setChecked(True)

        # connect signal / slot
        self._preProcessingWidget.sigConfChanged.connect(
            self._signalConfChangedPreProcessing)
        self._phaseWidget.sigConfChanged.connect(
            self._signalConfChangedPhase)
        self._reconstructionWidget.sigConfChanged.connect(
            self._signalConfChangedReconstruction)

        # expose API
        self.getSlices = self._reconstructionWidget.getSlices
        self.setOutputDir = self._outputWidget.setOutputDir

    def _signalConfChangedPhase(self, param):
        self.sigConfChanged.emit(self._phaseWidget.getStage().value, param)

    def _signalConfChangedPreProcessing(self, param):
        self.sigConfChanged.emit(self._preProcessingWidget.getStage().value, param)

    def _signalConfChangedReconstruction(self, param):
        self.sigConfChanged.emit(self._reconstructionWidget.getStage().value, param)

    def getConfiguration(self):
        config = {
            'preproc': self._preProcessingWidget.getConfiguration(),
            'reconstruction': self._reconstructionWidget.getConfiguration(),
            'tomwer_slices': self.getSlices(),
            'output': self._outputWidget.getConfiguration(),
        }
        if self._phaseGB.isChecked():
            config['phase'] = self._phaseWidget.getConfiguration()
        return config

    def setConfiguration(self, config):
        if 'preproc' in config:
            self._preProcessingWidget.setConfiguration(config['preproc'])
        if 'phase' in config:
            self._phaseWidget.setConfiguration(config['phase'])
            self._phaseGB.setChecked(True)
        else:
            self._phaseGB.setChecked(False)
        if 'reconstruction' in config:
            self._reconstructionWidget.setConfiguration(config['reconstruction'])
        if 'tomwer_slices' in config:
            self._reconstructionWidget.setSlices(config['tomwer_slices'])


if __name__ == '__main__':
    app = qt.QApplication([])
    widget = NabuConfiguration(None)
    widget.show()
    print(widget.getConfiguration())
    print(widget.getSlices())
    app.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


import os
import copy
try:
    from contextlib import AbstractContextManager
except ImportError:
    from tomwer.third_party.contextlib import AbstractContextManager
import datetime
from silx.utils.enum import Enum as _Enum
import tomwer.version
from tomwer.core.log import TomwerLogger
from tomwer.core.scan.scanbase import _TomwerBaseDock
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.process.baseprocess import (SingleProcess, _input_desc,
                                             _output_desc)
from nabu.resources.processconfig import ProcessConfig
from nabu.resources.dataset_analyzer import EDFDatasetAnalyzer, HDF5DatasetAnalyzer
from nabu.resources.dataset_analyzer import DatasetAnalyzer
from multiprocessing import Process
from nabu.io.config import generate_nabu_configfile
from nabu import version as nabu_version
from silx.io.dictdump import h5todict
import h5py
import sys

_logger = TomwerLogger(__name__)


NABU_FILE_PER_GROUP = 100

NABU_CONFIG_FILE_EXTENSION = '.cfg'


def run_reconstruction(scan: TomwerScanBase, config: dict,
                       dry_run: bool = False, local: bool = True) -> None:
    """
    call nabu for a reconstruction on scan with the given configuration

    :param TomwerScanBase scan: scan to reconstruct
    :param dict config: configuration to run the reconstruction
    :param bool dry_run: do we want to run dry
    :param bool local: do we want to run a local reconstruction
    """
    _logger.info('start reconstruction of %s' % scan.path)
    # if scan contains some center of position copy it to nabu
    if scan.axis_params is not None and scan.axis_params.value is not None:
        if 'reconstruction' in config:
            # move the cor value to the nabu reference
            cor_nabu_ref = scan.axis_params.value + scan.dim_1 // 2
            config['reconstruction']['rotation_axis_position'] = str(cor_nabu_ref)

    nabu_configurations = _interpret_tomwer_configuration(config, scan=scan)
    for nabu_configuration in nabu_configurations:
        config, slice_index = nabu_configuration
        _run_single_reconstruction(config=config, scan=scan, local=local,
                                   slice_index=slice_index, dry_run=dry_run)


class Nabu(SingleProcess):
    """Definition of the nabu reconstruction Single process"""
    inputs = [
        _input_desc(name="change recons params", type=_TomwerBaseDock,
                    handler='updateReconsParam',
                    doc='input with scan + reconstruction parameters'),
        _input_desc(name='data', type=TomwerScanBase, handler='pathReceived',
                    doc='scan path'),
    ]
    # Note : scanReady don't intend to find an 'octave_FT_params.h5' file at
    # the folder level.
    # But updateReconsParam should always have a .h5 file defined
    outputs = [_output_desc(name='data', type=TomwerScanBase, doc='scan path'), ]

    def __init__(self, *arg, **kwargs):
        SingleProcess.__init__(self, *arg, **kwargs)
        self._dry_run = False

    def process(self, scan=None):
        if scan is None:
            return None
        else:
            run_reconstruction(scan=scan, config=self.get_configuration(),
                               dry_run=self.dry_run)
            # register result
            entry = 'entry'
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry
            with scan.acquire_process_file_lock():
                self.register_process(process_file=scan.process_file,
                                      entry=entry,
                                      configuration=self.get_configuration(),
                                      results={},
                                      process_index=scan.pop_process_index(),
                                      overwrite=True)

    @staticmethod
    def program_name():
        return 'nabu'

    @staticmethod
    def program_version():
        return nabu_version

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @staticmethod
    def get_process_frm_process_file(process_file, entry):
        """
        Read informations regarding the nabu process save in the
        tomwer_process.h5 file

        :param process_file:
        :param entry:
        :return: dictionnary with the contain of the nabu process
        :rtype:dict
        """
        if entry is None:
            with h5py.File(process_file, 'r', swmr=True) as h5f:
                entries = Nabu._get_process_nodes(root_node=h5f,
                                                  process=Nabu)
                if len(entries) == 0:
                    _logger.info(
                        'unable to find a Axis process in %s' % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError('several entry found, entry should be '
                                     'specify')
                else:
                    entry = list(entries.keys())[0]
                    _logger.info('take %s as default entry' % entry)

        configuration_path = None
        res = {}

        with h5py.File(process_file, 'r', swmr=True) as h5f:
            nabu_nodes = Nabu._get_process_nodes(root_node=h5f[entry],
                                                 process=Nabu)
            index_to_path = {}
            for key, value in nabu_nodes.items():
                index_to_path[key] = key

            if len(nabu_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(list(nabu_nodes.keys()))[-1]
            last_process_dark = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug('several processing found for dark-ref,'
                             'take the last one: %s' % last_process_dark)

            for key_name in ('class_instance', 'date', 'program',
                             'sequence_index', 'version'):
                if key_name in h5f[last_process_dark]:
                    res[key_name] = h5f[last_process_dark][key_name][()]
            if 'configuration' in h5f[last_process_dark]:
                configuration_path = '/'.join((h5f[last_process_dark].name,
                                               'configuration'))

        if configuration_path is not None:
            res['configuration'] = h5todict(h5file=process_file,
                                            path=configuration_path)
        return res


def _interpret_tomwer_configuration(config: dict, scan: TomwerScanBase) -> tuple:
    """
    tomwer can 'mock' the nabu reconstruction to request more feature.
    Typical use case is that we can ask for reconstruction of several
    slices and not only the volume

    :param dict config: tomwer configuration for nabu
    :return: tuple of tuples (nabu configuration, is slice)
    """
    def get_nabu_config(config):
        nabu_config = copy.deepcopy(config)
        if 'tomwer_slices' in nabu_config:
            del nabu_config['tomwer_slices']
        return nabu_config

    if 'tomwer_slices' in config:
        slices = list(NabuSliceMode.getSlices(config['tomwer_slices'], scan))
    else:
        slices = []

    if 'phase' in config and 'delta_beta' in config['phase']:
        pag_dbs = _retrieve_lst_of_value_from_str(config['phase']['delta_beta'])
        if len(pag_dbs) == 0:
            pag_dbs = (None,)
    else:
        pag_dbs = (None,)

    # by default add the slice 'None' which is the slice for the volume
    slices.append(None)
    nabu_config = get_nabu_config(config=config)
    res = []
    for slice in slices:
        for pag_db in pag_dbs:
            local_config = copy.deepcopy(nabu_config)
            if slice is not None:
                local_config['reconstruction']['start_z'] = slice
                local_config['reconstruction']['end_z'] = slice + 1
            if pag_db is not None:
                local_config['phase']['delta_beta'] = str(pag_db)
            res.append((local_config, slice))
    return tuple(res)


def _get_file_basename_reconstruction(scan, slice_index, pag, db):
    """

    :param TomwerScanBase scan: scan reconstructed
    :param Union[None, int] slice_index: index of the slice reconstructed.
                                         if None, we want to reconstruct the
                                         entire volume
    :param bool pag: is it a paganin reconstruction
    :param int db: delta / beta parameter
    :return: basename of the file reconstructed (without any extension)
    """
    assert type(db) in (int, type(None))
    if slice_index is None:
        if pag:
            return '_'.join((os.path.basename(scan.path) + 'pag', 'db' + str(db).zfill(4)))
        else:
            return os.path.basename(scan.path)
    else:
        if pag:
            return '_'.join((os.path.basename(scan.path) + 'slice_pag',
                             str(slice_index).zfill(4),
                             'db' + str(db).zfill(4)))
        else:
            return '_'.join((os.path.basename(scan.path) + 'slice',
                             str(slice_index).zfill(4)))


def _launch_reconstruction(config_file, slice_index):
    from nabu.resources.processconfig import ProcessConfig
    proc = ProcessConfig(config_file)
    
    from nabu.app.fullfield_cuda import CudaFullFieldPipeline
    worker_process = CudaFullFieldPipeline(
        proc,
        sub_region=(None, None, slice_index, slice_index + 1),
    )
    worker_process.process_chunk()


def _run_single_reconstruction(scan, config, dry_run, slice_index, local,
                               use_cuda=True):
    """

    :param scan:
    :param config:
    :param dry_run:
    :param Union[None,int] slice_index: slice index to reconstruct
    :param local:
    :param use_cuda:
    :return:
    """
    dataset_analyzer = get_nabu_dataset_analyzer(scan=scan)
    config['dataset'] = get_nabu_dataset_desc(scan=scan)
    if local is True:
        resources_method = 'local'
    else:
        resources_method = 'slurm'
    config['resources'] = get_nabu_resources_desc(scan=scan, workers=1,
                                                  method=resources_method)
    config['about'] = get_nabu_about_desc(overwrite=True)

    def treateOutputConfig(_config):
        """
        - add or overwrite some parameters of the dictionary
        - create the output directory if does not exist
        """
        pag = False
        db = None
        if 'phase' in _config:
            if 'method' in _config['phase'] and _config['phase']['method'] != '':
                pag = True
                if 'delta_beta' in _config['phase']:
                    db = round(float(_config['phase']['delta_beta']))
        file_name = _get_file_basename_reconstruction(scan=scan,
                                                      slice_index=slice_index,
                                                      pag=pag,
                                                      db=db)
        if 'output' in config:
            _config['output']['file_prefix'] = os.path.basename(scan.path)
            _config['output']['file_prefix'] = file_name
            if _config['output']['location'] not in ('', None):
                # if user specify the location
                if not os.path.isdir(_config['output']['location']):
                    os.makedirs(_config['output']['location'])
            else:
                # otherwise default location will be the data root level
                _config['output']['location'] = scan.path
        return _config

    config = treateOutputConfig(config)
    # the policy is to save nabu .cfg file at the same location as the
    # reconstructed scan
    process_config = ProcessConfig(conf_dict=config,
                                   dataset_infos=dataset_analyzer)
    # write the .conf file
    conf_file = os.path.join(config['output']['location'],
                            config['output']['file_prefix'] + NABU_CONFIG_FILE_EXTENSION)
    _logger.info('create %s' % conf_file)

    # add some tomwer metadata and save the configuration
    # note: for now the section is ignored by nabu but shouldn't stay that way
    with _TomwerInfo(config) as config_to_dump:
        generate_nabu_configfile(fname=conf_file, config=config_to_dump,
                                 options_level='advanced')
        
    if slice_index is not None and dry_run is False and local:
        _logger.info('run nabu reconstruction for %s with %s'
                     '' % (scan.path, config))
        # need to be excuted in his own context
        process = Process(target=_launch_reconstruction, args=(conf_file, slice_index))
        process.start()
        process.join()


def get_nabu_dataset_analyzer(scan: TomwerScanBase) -> DatasetAnalyzer:
    """

    :param scan:
    :return:
    """
    if isinstance(scan, EDFTomoScan):
        return EDFDatasetAnalyzer(location=scan.path)
    elif isinstance(scan, HDF5TomoScan):
        assert os.path.exists(scan.master_file)
        assert os.path.isfile(scan.master_file)
        return HDF5DatasetAnalyzer(location=scan.master_file)
    else:
        raise TypeError('given scan type %s is not managed' % type(scan))


def get_nabu_dataset_desc(scan: TomwerScanBase) -> dict:
    """
    Create the descriptor for the outputs of scan for nabu

    :param TomwerScanBase scan:
    :return: nabu's description for the output
    """
    assert isinstance(scan, TomwerScanBase)
    if isinstance(scan, EDFTomoScan):
        return {
            'location': scan.path,
            'binning': 1,
            'binning_z': 1,
            'projections_subsampling': 1,
        }
    elif isinstance(scan, HDF5TomoScan):
        return {
            'location': scan.master_file,
            'binning': 1,
            'binning_z': 1,
            'projections_subsampling': 1,
        }
    else:
        raise ValueError('TomoBase type not recognized: ' + str(type(scan)))


def get_nabu_resources_desc(scan: TomwerScanBase, method, workers=1) -> dict:
    """
    Create the descriptor of nabu's resources

    :param TomwerScanBase scan:
    :param str method:
    :return: nabu's description of resources to be used
    """
    assert isinstance(scan, TomwerScanBase)
    res = {
        'method': method,
        'cpu_workers': workers,
        'queue': 'gpu',
        'memory_per_node': '90%',
        'threads_per_node': '100%',
        'walltime': '01:00:00',
    }
    return res


def get_nabu_about_desc(overwrite) -> dict:
    """
    Create the description for nabu's about

    :param self:
    :return:
    """
    return {
        'overwrite_results': str(bool(overwrite)),
    }


class _NabuMode(_Enum):
    FULL_FIELD = 'full field imaging'
    HALF_ACQ = 'half acquisition'


class _NabuStages(_Enum):
    INI = 'initialization'
    PRE = 'pre-processing'
    PHASE = 'phase'
    PROC = 'processing'
    POST = 'post-processing'

    @staticmethod
    def getStagesOrder():
        return (_NabuStages.INI, _NabuStages.PRE, _NabuStages.PHASE,
                _NabuStages.PROC, _NabuStages.POST)

    @staticmethod
    def getProcessEnum(stage):
        """Return the process Enum associated to the stage"""
        stage = _NabuStages.from_value(stage)
        if stage is _NabuStages.INI:
            raise NotImplementedError()
        elif stage is _NabuStages.PRE:
            return _NabuPreprocessing
        elif stage is _NabuStages.PHASE:
            return _NabuPhase
        elif stage is _NabuStages.PROC:
            return _NabuProcessing
        elif stage is _NabuStages.POST:
            return _NabuPostProcessing
        raise NotImplementedError()


class _NabuPreprocessing(_Enum):
    """Define all the preprocessing action possible and the order they
    are applied on"""
    FLAT_FIELD_NORMALIZATION = 'flat field normalization'
    CCD_FILTER = 'hot spot correction'

    @staticmethod
    def getPreProcessOrder():
        return (_NabuPreprocessing.FLAT_FIELD_NORMALIZATION,
                _NabuPreprocessing.CCD_FILTER)


class _NabuPhase(_Enum):
    """Define all the phase action possible and the order they
    are applied on"""
    PHASE = 'phase retrieval'
    UNSHARP_MASK = 'unsharp mask'
    LOGARITHM = 'logarithm'

    @staticmethod
    def getPreProcessOrder():
        return (_NabuPhase.PHASE, _NabuPhase.UNSHARP_MASK,
                _NabuPhase.LOGARITHM)


class _NabuProcessing(_Enum):
    """Define all the processing action possible"""
    RECONSTRUCTION = 'reconstruction'

    @staticmethod
    def getProcessOrder():
        return (_NabuProcessing.RECONSTRUCTION, )


class _ConfigurationLevel(_Enum):
    REQUIRED = 'required'
    OPTIONAL = 'optional'
    ADVANCED = 'advanced'

    def _get_num_value(self) -> int:
        if self is self.REQUIRED:
            return 0
        elif self is self.OPTIONAL:
            return 1
        elif self is self.ADVANCED:
            return 2

    def __le__(self, other):
        assert isinstance(other, _ConfigurationLevel)
        return self._get_num_value() <= other._get_num_value()


class _NabuPostProcessing(_Enum):
    """Define all the post processing action available"""
    SAVE_DATA = 'save'

    @staticmethod
    def getProcessOrder():
        return (_NabuPostProcessing.SAVE_DATA, )


class _NabuReconstructionMethods(_Enum):
    FBP = 'FBP'


class _NabuPhaseMethod(_Enum):
    """
    Nabu phase method
    """
    PAGANIN = 'Paganin'


class _NabuFBPFilterType(_Enum):
    RAMLAK = 'ramlak'


class _NabuPaddingType(_Enum):
    ZEROS = 'zeros'
    EDGES = 'edges'


def _retrieve_lst_of_value_from_str(my_string: str) -> tuple:
    """
    Return a list of value from a string like '12,23' or '(12, 23)',
    '[12;23]', '12;23' or with the pattern from:to:step like '0:10:1'
    :param str mystring:
    :return: list of single value
    """
    assert type(my_string) is str, 'work on string only'
    res = []
    my_string = my_string.replace('(', '')
    my_string = my_string.replace(')', '')
    my_string = my_string.replace('[', '')
    my_string = my_string.replace(']', '')
    if my_string.count(':') == 2:
        _from, _to, _step = my_string.split(':')
        _from, _to, _step = int(_from), int(_to), int(_step)
        if (_from > _to):
            tmp = _to
            _to = _from
            _from = tmp
        while _from <= _to:
            res.append(_from)
            _from += _step
        return tuple(res)
    else:
        vals = my_string.replace(' ', '')
        vals = vals.replace('_', '')
        vals = vals.replace(';', ',').split(',')
        for val in vals:
            try:
                res.append(int(val))
            except Exception:
                pass
        return tuple(res)


class NabuSliceMode(_Enum):
    MIDDLE = 'middle'
    OTHER = 'other'

    @staticmethod
    def getSlices(slices, scan) -> tuple:
        res = []
        try:
            mode = NabuSliceMode.from_value(slices)
        except ValueError:
            try:
                res = _retrieve_lst_of_value_from_str(slices)
            except Exception:
                pass
        else:
            if mode == mode.MIDDLE:
                n_slice = scan.dim_2 or 2048
                res.append(n_slice // 2)
            else:
                raise ValueError('there should be only two ways of defining '
                                 'slices: middle one or other, by giving '
                                 'an unique value or a list or a tuple')
        return tuple(res)


class _TomwerInfo(AbstractContextManager):
    """Simple context manager to add tomwer metadata to a dict before
    writing it"""
    def __init__(self, config_dict):
        self.config = config_dict

    def __enter__(self):
        self.config['other'] = {
            'tomwer_version': tomwer.version.version,
            'date': datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
        }
        return self.config

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.config['other']['tomwer_version']
        del self.config['other']['date']



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/07/2017"


import os
import shutil

from tomwer.core.process.baseprocess import SingleProcess, _input_desc, \
    _output_desc
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.settings import get_lbsram_path, get_dest_path
from tomwer.core.signal import Signal
from tomwer.core.utils import logconfig, rebaseParFile
from tomwer.core.log import TomwerLogger
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.hdf5scan import HDF5TomoScan
import tomwer.version

logger = TomwerLogger(__name__)

try:
    from tomwer.synctools.rsyncmanager import RSyncManager
except ImportError:
    logger.warning('rsyncmanager not available')
    has_rsync = False
else:
    has_rsync = True


class FolderTransfert(SingleProcess):
    """Manage the copy of scan.

    .. warning : the destination directory is find out from the file system
                 if /lbsramxxx exists for example...
                 In the case we couldn't found the output directory then we
                 will ask for the user to set it.
    """

    inputs = [_input_desc(name='data', type=TomwerScanBase, handler='process',
                          doc='scan path'), ]
    outputs = [_output_desc(name='data', type=TomwerScanBase, doc='scan path')]

    scanready = Signal(TomwerScanBase)

    def __init__(self):
        SingleProcess.__init__(self)
        self.turn_off_print = False
        self._destDir = None
        """
        output directory if forced. By default based by the env variable
        'TARGET_OUTPUT_FOLDER' if exists, else set to '/data/visitor'
         """
        self._copying = False
        self._block = False

    def setProperties(self, properties):
        # No properties stored for now
        if 'dest_dir' in properties:
            self.setDestDir(properties['dest_dir'])

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return 'data transfer'

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return 'transfer data from a folder to another'

    def _getDefaultOutputDir(self):
        """Return the default output dir based on the computer setup
        """
        if 'TARGET_OUTPUT_FOLDER' in os.environ:
            return os.environ['TARGET_OUTPUT_FOLDER']

        if os.path.isdir('/data/visitor'):
            return '/data/visitor'

        return ''

    def _process_edf_scan(self, scan, move=False, force=True, noRsync=False):
        assert isinstance(scan, EDFTomoScan)
        outputdir = self.getDestinationDir(scan.path)
        if outputdir is None:
            return

        self._pretransfertOperations(scan.path, outputdir)
        # as we are in the workflow we want this function to be bloking.
        # so we will not used a thread for folder synchronization
        # for now rsync is not delaing with force option
        if not has_rsync or noRsync is True or RSyncManager().has_rsync() is False:
            logger.info("Can't use rsync, copying files")
            try:
                if move is True:
                    self._moveFiles(scanPath=scan.path,
                                    outputdir=os.path.dirname(outputdir),
                                    force=force)
                else:
                    self._copyFiles(scanPath=scan.path,
                                    outputdir=outputdir,
                                    force=force)
            except shutil.Error as e:
                raise e
            else:
                if isinstance(scan, HDF5TomoScan):
                    _entry = scan.entry
                else:
                    _entry = None
                output_scan = scan._deduce_transfert_scan(outputdir)
                self.__noticeTransfertSuccess(input_scan=scan,
                                              output_scan=output_scan)
        else:
            source = scan.path
            if not source.endswith(os.path.sep):
                source = source + os.path.sep
            target = outputdir

            if not target.endswith(os.path.sep):
                target = target + os.path.sep

            self._signalCopying(scanID=source, outputdir=target)
            output_scan = scan._deduce_transfert_scan(outputdir)

            RSyncManager().sync_file(source=source,
                                     target=target,
                                     wait=self._block,
                                     delete=True,
                                     callback=self.__noticeTransfertSuccess,
                                     callback_parameters=(scan, output_scan),
                                     rights=777)

        self.register_output(key='data', value=outputdir)

    def _get_hdf5_dst_scan(self, bliss_scan_folder_path):
        if self._destDir is not None:
            rel_path = os.path.join(*bliss_scan_folder_path.split(os.sep)[-2:])
            return os.path.dirname(os.path.join(self._destDir, rel_path))
        # try to get outputdir from spec
        scanIDPath = os.path.abspath(bliss_scan_folder_path)
        return self._getOutputDirSpec() or os.path.join(*self._getOutputDirLBS(scanIDPath).split(os.sep)[:-1])

    def _get_hdf5_sample_file_or_nx_dst(self, bliss_sample_file):
        if self._destDir is not None:
            rel_path = os.path.join(*bliss_sample_file.split(os.sep)[-2:])
            return os.path.join(self._destDir, rel_path)
        # try to get outputdir from spec
        bliss_sample_file = os.path.abspath(bliss_sample_file)
        return self._getOutputDirSpec() or self._getOutputDirLBS(bliss_sample_file)

    def _get_master_sample_file_dst(self, master_sample_file):
        if self._destDir is not None:
            rel_path = os.path.join(*master_sample_file.split(os.sep)[-2:])
            return os.path.join(self._destDir, rel_path)
        # try to get outputdir from spec
        master_sample_file = os.path.abspath(master_sample_file)
        return self._getOutputDirSpec() or self._getOutputDirLBS(master_sample_file)

    def _get_hdf5_proposal_file_dst(self, bliss_proposal_file):
        if self._destDir is not None:
            return os.path.join(self._destDir, os.path.split(bliss_proposal_file)[-1])
        # try to get outputdir from spec
        bliss_sample_file = os.path.abspath(bliss_proposal_file)
        return self._getOutputDirSpec() or self._getOutputDirLBS(bliss_sample_file)

    def _process_hdf5_scan(self, scan):
        assert isinstance(scan, HDF5TomoScan)
        from tomwer.core.process.datalistener.datalistener import DataListener

        files_sources = []
        files_dest = []
        delete_opt = []

        if os.path.exists(scan.process_file):
            source_scans = DataListener.get_source_scans(
                process_file=scan.process_file,
                entry=scan.entry)
        else:
            source_scans = None
        # manage scan folder
        if source_scans is not None:
            for source_scan in source_scans:
                files_sources.append(source_scan)
                files_dest.append(self._get_hdf5_dst_scan(bliss_scan_folder_path=source_scan))
                delete_opt.append(True)

        # manage proposal file
        if os.path.exists(scan.process_file):
            proposal_file = DataListener.get_proposal_file(process_file=scan.process_file,
                                                           entry=scan.entry)
        else:
            proposal_file = None
        if proposal_file is not None:
            files_sources.append(proposal_file)
            files_dest.append(self._get_hdf5_proposal_file_dst(proposal_file))
            delete_opt.append(False)

        # manage sample file
        if os.path.exists(scan.process_file):
            sample_file = DataListener.get_sample_file(process_file=scan.process_file,
                                                       entry=scan.entry)

        else:
            sample_file = None
        if sample_file is not None:
            files_sources.append(sample_file)
            files_dest.append(self._get_hdf5_sample_file_or_nx_dst(sample_file))
            delete_opt.append(False)

        # manage saving file
        if os.path.exists(scan.process_file):
            mst_sample_file = DataListener.get_master_sample_file(
                process_file=scan.process_file,
                entry=scan.entry)
            print('saving file found: {}'.format(mst_sample_file))
        else:
            mst_sample_file = None
        if mst_sample_file is not None:
            print('saving file destination will be {}'.format(self._get_master_sample_file_dst(mst_sample_file)))
            files_sources.append(mst_sample_file)
            files_dest.append(
                self._get_master_sample_file_dst(mst_sample_file))
            delete_opt.append(False)

        # manage .nx file
        if os.path.exists(scan.master_file):
            files_sources.append(scan.master_file)
            new_nx_file = self._get_hdf5_sample_file_or_nx_dst(scan.master_file)
            files_dest.append(new_nx_file)
            delete_opt.append(True)
            output_scan = HDF5TomoScan(scan=new_nx_file, entry=scan.entry)
        else:
            output_scan = None
        RSyncManager().sync_files(sources=files_sources,
                                  targets=files_dest,
                                  wait=self._block,
                                  delete=delete_opt,
                                  callback=self.__noticeTransfertSuccess,
                                  callback_parameters=(scan, output_scan))

    def process(self, scan, move=False, force=False, noRsync=False):
        """Launch the process process

        :param scan: the path to the file we want to move/process
        :type scan: :class:`.TomoBase`
        :param move: if True, directly move the files. Otherwise copy the files
        :param force: if True then force the copy even if the file/folder already
            exists
        :param bool noRSync: True if we wan't do sue shutil instead of rsync.
        """
        if scan is None:
            return

        _scan = scan
        if type(_scan) is dict:
            _scan = ScanFactory.create_scan_object_frm_dict(scan)

        assert isinstance(_scan, TomwerScanBase)
        assert move in (True, False)

        logger.info('synchronisation with scanPath')
        if isinstance(scan, EDFTomoScan):
            self._process_edf_scan(scan=scan, move=move, force=force,
                                   noRsync=noRsync)
        elif isinstance(scan, HDF5TomoScan):
            if move is True:
                raise NotImplementedError('move option not implemented')
            if noRsync is True:
                raise NotImplementedError('noRsync option not implemented')

            self._process_hdf5_scan(scan=scan)
        else:
            raise TypeError('Other scan than EDF or HDF5 are not managed')

        if self._return_dict:
            return _scan.to_dict()
        else:
            return _scan

    def _pretransfertOperations(self, scanfolder, outputdir):
        """Operation to be run before making the transfert of the scan"""
        self._updateParFiles(scanfolder, outputdir)

    def _updateParFiles(self, scanfolder, outputdir):
        """Update all path contained in the .par files to fit the new outpudir
        """
        for _file in os.listdir(scanfolder):
            if _file.lower().endswith('.par'):
                rebaseParFile(os.path.join(scanfolder, _file),
                              oldfolder=scanfolder, newfolder=outputdir)

    def __noticeTransfertSuccess(self, input_scan, output_scan):
        self._signalCopySucceed()

        results = {'input_scan': str(input_scan),
                   'output_scan': str(output_scan)}
        entry = 'entry'
        if isinstance(output_scan, HDF5TomoScan):
            entry = output_scan.entry
        self.register_process(process_file=output_scan.process_file,
                              entry=entry, results=results,
                              process_index=output_scan.pop_process_index(),
                              configuration=None)
        logger.processEnded('transfert succeed',
                            extra={logconfig.DOC_TITLE: self._scheme_title,
                                   logconfig.FROM: str(input_scan),
                                   logconfig.TO: str(output_scan)})
        self.signalTransfertOk(input_scan=input_scan, output_scan=output_scan)

    def signalTransfertOk(self, input_scan, output_scan):
        if input_scan is None or output_scan is None:
            return
        assert isinstance(input_scan, TomwerScanBase)
        assert isinstance(output_scan, TomwerScanBase)
        self.scanready.emit(output_scan)

    def _copyFiles(self, scanPath, outputdir, force):
        """Copying files and removing them"""
        assert type(scanPath) is str
        assert type(outputdir) is str
        assert(os.path.isdir(scanPath))
#        if force is False:
#            assert(os.path.isdir(outputdir))
        # create the destination dir
        if not os.path.isdir(outputdir):
            def createDirAndTopDir(_dir):
                if not os.path.isdir(os.path.dirname(_dir)):
                    createDirAndTopDir(os.path.dirname(_dir))
                os.mkdir(_dir)
            createDirAndTopDir(outputdir)
        # we can't copy directly the top folder because he is already existing
        for f in os.listdir(scanPath):
            file = os.path.join(scanPath, f)
            fileDest = os.path.join(outputdir, f)
            if force is True:
                if os.path.isdir(fileDest):
                    shutil.rmtree(fileDest)
                if os.path.isfile(fileDest):
                    os.remove(fileDest)
            if os.path.exists(fileDest):
                raise FileExistsError(fileDest, 'already exists')
            if os.path.isdir(file):
                shutil.copytree(src=file, dst=fileDest)
            else:
                shutil.copy2(src=file, dst=fileDest)

        info = 'Removing directory at %s' % scanPath
        logger.info(info)
        shutil.rmtree(scanPath)
        info = 'sucessfuly removed file at %s !!!' % scanPath
        logger.info(info)

    def _moveFiles(self, scanPath, outputdir, force):
        """Function simply moving files"""
        assert(os.path.isdir(scanPath))
        if force is False:
            assert(os.path.isdir(outputdir))

        logger.debug('synchronisation with scanPath',
                     extra={logconfig.DOC_TITLE: self._scheme_title})

        target = os.path.join(outputdir, os.path.basename(scanPath))
        if force is True and os.path.isdir(target):
            shutil.rmtree(target)
        shutil.move(scanPath, outputdir)

    def _requestFolder(self):
        out = None
        while(out is None):
            out = input('please give the output directory : \n')
            if not os.path.isdir(out):
                warning = 'given path ' + out
                warning += ' is not a directory, please give a valid directory'
                logger.warning(warning)
                out = None
        return out

    def _getOutputDirSpec(self):
        return None

    def _getOutputDirLBS(self, scanPath):
        if scanPath.startswith(get_lbsram_path()):
            return scanPath.replace(get_lbsram_path(), get_dest_path(), 1)
        else:
            return None

    def getDestinationDir(self, scanPath):
        """Return the destination directory. The destination directory is the
        root directory"""
        if self._destDir is not None:
            return os.path.join(self._destDir, os.path.basename(scanPath))

        # try to get outputdir from spec
        scanIDPath = os.path.abspath(scanPath)

        outputdir = self._getOutputDirSpec() or self._getOutputDirLBS(scanIDPath)
        if outputdir is None:
            outputdir = self._requestFolder()

        return outputdir

    def setDestDir(self, dist):
        """Force the outpudir to dist.

        :param str dist: path to the folder. If None remove force behavior
        """
        self._destDir = dist
        if self._destDir is not None and os.path.isdir(self._destDir):
            logger.warning('Given path %s is not a directory' % self._destDir)

    # some function to print the output in the terminal #

    def _signalCopying(self, scanID, outputdir):
        self._copying = True
        if self.turn_off_print is False:
            print('######################################')
            print('###')
            print('###')
            print('### copying files ', scanID, " to ", outputdir)
            print('### ...')

        info = 'start moving folder from %s to %s' % (scanID, outputdir)
        logger.info(info,
                    extra={logconfig.DOC_TITLE: self._scheme_title})

    def _signalCopyFailed(self):
        self._copying = False
        if self.turn_off_print is False:
            print('###')
            print('### copy failed')
            print('###')
            print('######################################')

    def _signalCopySucceed(self):
        self._copying = False
        if self.turn_off_print is False:
            print('###')
            print('### copy succeeded')
            print('###')
            print('######################################')

    def isCopying(self):
        """

        :return: True if the folder transfert is actually doing a copy
        """
        return self._copying

    def setForceSync(self, b):
        """

        :param bool b: if True then folderTransfert will wait until transfert
            is done to be released. Otherwise will launch a 'free' thread wich
            will notice transfert end later.
        """
        self._block = b

    def getOutput(self, scan):
        """
        
        :param scan: 
        :return: 
        """
        return os.path.join(self.getDestinationDir(scan), os.path.basename(scan))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/07/2017"


import h5py
import os
from tomwer.core import settings
import tomwer.version
from tomwer.core.process.baseprocess import SingleProcess, _output_desc
from tomwer.core.log import TomwerLogger
from tomwer.core.scan.scanbase import TomwerScanBase
from .rpcserver import DataListenerThread
from nxtomomill import converter as nxtomomill_converter
from tomwer.core.scan.hdf5scan import HDF5TomoScan

_logger = TomwerLogger(__name__)


def nxtomomill_input_callback(entry, desc):
    if entry =='energy':
        _logger.warning("Energy is missing {}. set to {}".format(entry, 19.0))
        return 19.0
    else:
        _logger.warning("missing {}. Won't be set".format(entry))
        return None


class DataListener(SingleProcess):
    """
    class able to connect with a redis database.
    During an acquisition at esrf the redis database expose the scan under
    acquisition. This allow use to know when a scan is finished.

    In order to insure connection the machine need to know where is located
    the 'bacon server'
    (see https://bliss.gitlab-pages.esrf.fr/bliss/master/bliss_data_life.html).
    This is why the 'BEACON_HOST' environment variable should be defined.

    For example for id19 - lbs 191 we have:
    export BEACON_HOST="europa"
    On id19 - europa we would have
    export BEACON_HOST="localhost"
    """

    outputs = [
        _output_desc(name='data', type=TomwerScanBase, doc='scan path'),
    ]

    TIMOUT_READ_FILE = 10
    "When the event 'scan_ended' is received all data might not have been write" \
    " yet"

    def __init__(self):
        SingleProcess.__init__(self)
        self._host = settings.JSON_RPC_HOST
        # if host is None then use hostname
        if self._host is None:
            import socket
            self._host = socket.gethostname()
        self._port = settings.JSON_RPC_PORT
        self._listening_thread = None

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return 'data listener'

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        'convert bliss sequence with a proposal to a NXTomo nexus file'

    @property
    def port(self):
        return self.port

    @property
    def host(self):
        return self.host

    def get_listening_thread(self):
        """
        return the listening thread to ear at tango server
        """
        if self._listening_thread is None:
            self._listening_thread = self.create_listening_thread()
        return self._listening_thread

    def create_listening_thread(self):
        """
        Procedure to create the listening thread
        :return: listening thread
        """
        return DataListenerThread(host=self.host, port=self.port)

    def delete_listening_thread(self):
        """
        Procedure to delete the listening thread
        """
        if hasattr(self, '_listening_thread'):
            if self._listening_thread is not None:
                self._listening_thread.stop()
                self._listening_thread.join(5)
            del self._listening_thread
        self._listening_thread = None

    def setProperties(self, properties):
        pass

    def process(self, scan=None):
        pass

    def process_sample_file(self, sample_file: str, entry: str,
                            proposal_file: str, master_sample_file: str) -> None:
        """

        :param sample_file: file to be converted
        :param entry: entry in the tango .h5 file to be converter
        :return: tuple (output file, output file entry). Both are set to None
                 if conversion fails.
        """
        if not os.path.isfile(sample_file):
            raise ValueError('Given file {} does not exists.'.format(sample_file))
        master_file, entry_master_file = self.convert(bliss_file=sample_file,
                                                      entry=entry)
        if master_file is not None and entry_master_file is not None:
            scan = HDF5TomoScan(scan=master_file, entry=entry_master_file)

            # register process.
            # also store the source 'scan' folder. This information will
            # be needed by the datatransfert if some given later.
            source_scans = DataListener._get_scan_sources(sample_file, entry)
            self.register_process(process_file=scan.process_file,
                                  entry=scan.entry,
                                  results={'output_file': scan.master_file,
                                           'entry': scan.entry},
                                  configuration={'sample_file': sample_file,
                                                 'entry': entry,
                                                 'source_scans': source_scans,
                                                 'file_proposal': proposal_file,
                                                 'master_sample_file': master_sample_file,
                                                 },
                                  process_index=scan.pop_process_index(),
                                  overwrite=True)

            self._signal_scan_ready(scan)

    @staticmethod
    def _get_scan_sources(bliss_file, entry) -> list:
        """Return the list of scans dir for this bliss_file / entry"""

        def get_scan_indexes():
            with h5py.File(bliss_file, 'r') as h5f:
                entry_node = h5f[entry]
                if nxtomomill_converter._Acquisition._SCAN_NUMBER_PATH in entry_node:
                    return entry_node[nxtomomill_converter._Acquisition._SCAN_NUMBER_PATH][()]
                else:
                    raise ValueError('Unable to find scan number indexes')
        scans_indexes = get_scan_indexes()
        res = []
        for scans_index in scans_indexes:
            # Bad hack
            # TODO: we should find a better way to retrieve source. From
            # virtual dataset I don't see it. Maybe from symbolic link.
            scan_folder_name = 'scan' + str(scans_index).zfill(4)
            scan_folder_name = os.path.join(os.path.dirname(bliss_file),
                                            scan_folder_name)
            res.append(scan_folder_name)
        return res

    @staticmethod
    def get_proposal_file(process_file, entry):
        """Return the proposal file of the experimentation if registred by the
        data listener"""
        if entry is None:
            with h5py.File(process_file, 'r', swmr=True) as h5f:
                entries = SingleProcess._get_process_nodes(root_node=h5f,
                                                           process=DataListener)
                if len(entries) == 0:
                    _logger.info(
                        'unable to find a DarkRef process in %s' % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError('several entry found, entry should be '
                                     'specify')
                else:
                    entry = list(entries.keys())[0]
                    _logger.info('take %s as default entry' % entry)

        with h5py.File(process_file, 'r', swmr=True) as h5f:
            dl_nodes = SingleProcess._get_process_nodes(root_node=h5f[entry],
                                                        process=DataListener)
            index_to_path = {}
            for key, value in dl_nodes.items():
                index_to_path[key] = key

            if len(dl_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.keys())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug('several processing found for data listener,'
                             'take the last one: %s' % last_process_dl)

            if 'configuration' in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]['configuration']
                if 'file_proposal' in results_node.keys():
                    try:
                        fp = results_node['file_proposal'][()].decode('UTF-8')
                    except AttributeError:
                        return None
                    else:
                        return fp
                else:
                    return None
            return None

    # TODO the 3 static functions get_proposal_file, get_sample_file and
    # get_source_scans are sharing a lot of source code and should
    # be 'concatenate'
    @staticmethod
    def get_sample_file(process_file, entry):
        """Return the proposal file of the experimentation if registred by the
        data listener"""
        if entry is None:
            with h5py.File(process_file, 'r', swmr=True) as h5f:
                entries = SingleProcess._get_process_nodes(root_node=h5f,
                                                           process=DataListener)
                if len(entries) == 0:
                    _logger.info(
                        'unable to find a DarkRef process in %s' % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError('several entry found, entry should be '
                                     'specify')
                else:
                    entry = list(entries.keys())[0]
                    _logger.info('take %s as default entry' % entry)

        with h5py.File(process_file, 'r', swmr=True) as h5f:
            dl_nodes = SingleProcess._get_process_nodes(root_node=h5f[entry],
                                                        process=DataListener)
            index_to_path = {}
            for key, value in dl_nodes.items():
                index_to_path[key] = key

            if len(dl_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.keys())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug('several processing found for data listener,'
                             'take the last one: %s' % last_process_dl)

            if 'configuration' in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]['configuration']
                if 'sample_file' in results_node.keys():
                    return results_node['sample_file'][()].decode('UTF-8')
                else:
                    return None
            return None

    @staticmethod
    def get_master_sample_file(process_file, entry):
        """Return the proposal file of the experimentation if registred by the
        data listener"""
        if entry is None:
            with h5py.File(process_file, 'r', swmr=True) as h5f:
                entries = SingleProcess._get_process_nodes(root_node=h5f,
                                                           process=DataListener)
                if len(entries) == 0:
                    _logger.info(
                        'unable to find a DarkRef process in %s' % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError('several entry found, entry should be '
                                     'specify')
                else:
                    entry = list(entries.keys())[0]
                    _logger.info('take %s as default entry' % entry)

        with h5py.File(process_file, 'r', swmr=True) as h5f:
            dl_nodes = SingleProcess._get_process_nodes(root_node=h5f[entry],
                                                        process=DataListener)
            index_to_path = {}
            for key, value in dl_nodes.items():
                index_to_path[key] = key

            if len(dl_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.keys())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug('several processing found for data listener,'
                             'take the last one: %s' % last_process_dl)

            if 'configuration' in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]['configuration']
                if 'master_sample_file' in results_node.keys():
                    return results_node['master_sample_file'][()].decode('UTF-8')
                else:
                    return None
            return None

    @staticmethod
    def get_source_scans(process_file, entry):
        """Return the list of 'bliss scan directory' created for holding this
        specific sequence data
        """
        if entry is None:
            with h5py.File(process_file, 'r', swmr=True) as h5f:
                entries = SingleProcess._get_process_nodes(root_node=h5f,
                                                           process=DataListener)
                if len(entries) == 0:
                    _logger.info(
                        'unable to find a DarkRef process in %s' % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError('several entry found, entry should be '
                                     'specify')
                else:
                    entry = list(entries.keys())[0]
                    _logger.info('take %s as default entry' % entry)

        with h5py.File(process_file, 'r', swmr=True) as h5f:
            dl_nodes = SingleProcess._get_process_nodes(root_node=h5f[entry],
                                                           process=DataListener)
            index_to_path = {}
            for key, value in dl_nodes.items():
                index_to_path[key] = key

            if len(dl_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.keys())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug('several processing found for data listener,'
                             'take the last one: %s' % last_process_dl)

            if 'configuration' in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]['configuration']
                if 'source_scans' in results_node.keys():
                    tmp_res = results_node['source_scans'][()]
                    res = [mystr.decode('UTF-8') for mystr in tmp_res]
                    return res
                else:
                    return None
            return None

    def convert(self, bliss_file: str, entry: str) -> tuple:
        """

        :param bliss_file: file to be converted
        :param entry: entry in the tango .h5 file to be converter
        :return: tuple (output file, output file entry). Both are set to None
                 if conversion fails.
        """
        output_file_name = os.path.basename(bliss_file)
        output_file_name = os.path.splitext(output_file_name)[0]
        output_file_name = '_'.join((output_file_name, entry.replace('.', '_').replace(':', '_') + '.nx'))
        output_file_path = os.path.join(os.path.dirname(bliss_file), output_file_name)

        if os.path.exists(output_file_path):
            if not self._ask_user_for_overwritting(output_file_path):
                return None, None

        if entry.startswith('/'):
            entries = (entry, )
        else:
            entries = ('/' + entry,)

        # work around: we need to wait a bit before converting the file
        # otherwise it looks like the file might not be ended to be
        # write
        def sequence_is_finished():
            try:
                with h5py.File(bliss_file, 'r') as h5f:
                    end_scan_path = '/'.join((entry, 'end_time'))
                    return end_scan_path in h5f
            except Exception:
                return False

        timeout = self.TIMOUT_READ_FILE
        import time
        while timeout > 0 and not sequence_is_finished():
            timeout = timeout - 0.2
            time.sleep(0.2)

        if timeout <= 0:
            _logger.error('unable to access {}@{}. (Write never ended)'.format(entry, bliss_file))
            return None, None
        # one more delay to insure can read it
        time.sleep(4)
        try:
            conv = nxtomomill_converter.h5_to_nx(input_file_path=bliss_file,
                                                 output_file=output_file_path,
                                                 entries=entries,
                                                 single_file=True,
                                                 ask_before_overwrite=False,
                                                 request_input=True,
                                                 file_extension='.nx',
                                                 input_callback=nxtomomill_input_callback)
        except Exception as e:
            _logger.error(
                'Fail to convert from tango file: %s to NXTomo.'
                'Conversion error is: %s' % (bliss_file, e))
            return None, None
        else:
            assert len(conv) == 1
            return conv[0]

    def _ask_user_for_overwritting(self, file_path):
        res = None
        while res not in ('Y', 'n'):
            res = input("The process will overwrite %s. Do you agree ? (Y/n)"
                        "" % file_path )

        return res == 'Y'

    def _signal_scan_ready(self, scan):
        assert isinstance(scan, HDF5TomoScan)
        pass

    def activate(self, activate=True):
        """
        activate or deactivate the thread. When deactivate call join and
        delete the thread

        :param bool activate:
        """
        if activate:
            if self._listening_thread is not None:
                _logger.info('listening is already activate')
            else:
                assert isinstance(self, DataListener)
                self.get_listening_thread().start()
        else:
            if self._listening_thread is None:
                return
            else:
                self.delete_listening_thread()

    def is_active(self):
        if self._listening_thread is None:
            return False
        else:
            return True



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to axis calculation using sinogram
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/10/2019"


from silx.gui import qt
from .radioaxis import RadioAxisWindow
from .sinogramaxis import SinogramAxisWindow
from ...utils.scandescription import ScanNameLabel
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.synctools.axis import QAxisRP
import logging

_logger = logging.getLogger(__file__)


class _AxisTypeSelection(qt.QGroupBox):
    sigSelectionChanged = qt.Signal(str)
    """Signal emitted when the selection changed. Value can be `sinogram` or
    `radio`"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, parent=parent)
        self.setTitle('compute center of rotation from')
        self.setLayout(qt.QHBoxLayout())
        self._radioRB = qt.QRadioButton('radios', parent=self)
        self.layout().addWidget(self._radioRB)

        self._sinogramRB = qt.QRadioButton('sinogram', parent=self)
        self.layout().addWidget(self._sinogramRB)

        # Signal / Slot connections
        self._radioRB.toggled.connect(self._selectionChanged)
        self._sinogramRB.toggled.connect(self._selectionChanged)

    def getSelection(self):
        if self._radioRB.isChecked():
            return 'radio'
        else:
            return 'sinogram'

    def setSelection(self, selection):
        if selection == 'radio':
            self._radioRB.setChecked(True)
        elif selection == 'sinogram':
            self._sinogramRB.setChecked(True)
        else:
            raise ValueError('invalid selection given')

    def _selectionChanged(self, *args, **kwargs):
        self.sigSelectionChanged.emit(self.getSelection())


class AxisWindow(qt.QMainWindow):
    """Main widget for the axis calculation"""

    sigComputationRequested = qt.Signal()
    """signal emitted when a computation is requested"""

    sigApply = qt.Signal()
    """signal emitted when the axis reconstruction parameters are validated"""

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    sigMethodChanged = qt.Signal(str)
    """signal emitted when the method for computation change"""

    def __init__(self, axis_params, parent=None):
        qt.QMainWindow.__init__(self, parent=parent)
        self._mainWidget = qt.QWidget(self)
        self._mainWidget.setLayout(qt.QVBoxLayout())
        self.setCentralWidget(self._mainWidget)

        self._axis_params = axis_params

        # add scan name
        self._scan_label = ScanNameLabel(parent=self)
        self._mainWidget.layout().addWidget(self._scan_label)

        # add selection
        self._selectionGB = _AxisTypeSelection(parent=self)
        self._mainWidget.layout().addWidget(self._selectionGB)

        # add widget for radio and sinogram axis
        self._axisWidget = _AxisWidget(parent=self, axis_params=axis_params)
        self._mainWidget.layout().addWidget(self._axisWidget)

        self.setCentralWidget(self._mainWidget)

        # set up configuration
        if axis_params.use_sinogram:
            selection = 'sinogram'
        else:
            selection = 'radio'
        self._selectionGB.setSelection(selection=selection)
        self._selectionGB._selectionChanged(self._selectionGB.getSelection())

        # connect signal / slots
        self._axisWidget.sigValidateRequest.connect(self._repeatValidateRequest)
        self._axisWidget.sigComputationRequested.connect(self._computationRequested)
        self._selectionGB.sigSelectionChanged.connect(self._axisWidget._axisTypeChanged)
        self._axisWidget.sigSinogramReady.connect(self._computationReady)

        # expose API
        self.setSelection = self._selectionGB.setSelection
        self.getSelection = self._selectionGB.getSelection
        self.hideLockButton = self._axisWidget.hideLockButton
        self.hideApplyButton = self._axisWidget.hideApplyButton
        self.setReconsParams = self._axisWidget.setReconsParams
        self.getPlotWidget = self._axisWidget.getPlotWidget
        #self.setLocked = self._axisWidget.setLocked
        self.setModeLock = self._axisWidget.setModeLock
        self.getAxis = self._axisWidget.getAxis
        self.isModeLock = self._axisWidget.isModeLock
        self.isValueLock = self._axisWidget.isValueLock

    def _disableForProcessing(self, *args, **kwargs):
        self._mainWidget.setEnabled(False)
        self._axisWidget.setEnabled(False)

    def _enableForProcessing(self, *args, **kwargs):
        self._mainWidget.setEnabled(True)
        self._axisWidget.setEnabled(True)

    def setScan(self, scan):
        """
        set the gui for this scan

        :param TomoBase scan:
        """
        self._scan_label.setScan(scan=scan)
        self._axisWidget.setScan(scan=scan)
        self._enableSinogramOpt(scan.scan_range == 360)

    def _enableSinogramOpt(self, b):
        if self._selectionGB.getSelection() == 'sinogram' and not b:
            change_selection_to_radio = True
        else:
            change_selection_to_radio = False
        self._selectionGB._sinogramRB.setEnabled(b)
        self._axisWidget._enableSinogram(b)
        if change_selection_to_radio:
            self._selectionGB.setSelection('radio')

    def _computationRequested(self):
        if self.getSelection() == 'sinogram':
            # for sinogram we need first to load some 'raw materials': the sinogram
            self._axisWidget._loadCurrentSinogram()
        else:
            self._computationReady()

    def _computationReady(self):
        self.sigComputationRequested.emit()

    def _repeatValidateRequest(self):
        self.sigApply.emit()

    def setPosition(self, frm: float, value: float):
        self._axisWidget.setPosition(value=value)


class _AxisWidget(qt.QMainWindow):
    """
    Control which plot and which settings widget is displayed
    """

    sigComputationRequested = qt.Signal()

    sigValidateRequest = qt.Signal()

    sigLockCORValueChanged = qt.Signal(bool)
    """bool: True if locked"""

    sigModeChanged = qt.Signal(str)

    sigSinogramReady = qt.Signal()

    def __init__(self, parent, axis_params):
        qt.QMainWindow.__init__(self, parent)
        self._settingsWidget = None
        self.setDockOptions(qt.QMainWindow.AnimatedDocks)

        self._settingsDockWidget = qt.QDockWidget(parent=self)
        self._settingsDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._settingsDockWidget)

        self._controlWidget = _ControlWidget(parent=self)
        self._controlWidget.setSizePolicy(qt.QSizePolicy.Minimum,
                                          qt.QSizePolicy.Minimum)
        self._controlDockWidget = qt.QDockWidget(parent=self)
        self._controlDockWidget.setWidget(self._controlWidget)
        self._controlDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._controlDockWidget)

        self._axis_params = axis_params
        self._radioAxis = RadioAxisWindow(parent=self, axis=axis_params)
        self._sinogramAxis = SinogramAxisWindow(parent=self, axis=axis_params)
        self._plotWidgets = qt.QWidget(self)
        self._plotWidgets.setLayout(qt.QVBoxLayout())
        self.setCentralWidget(self._plotWidgets)

        # connect signal / slots
        self._controlWidget.sigComputationRequest.connect(self._repeatComputationRequest)
        self._controlWidget.sigValidateRequest.connect(self._repeatValidateRequest)
        self._controlWidget.sigLockCORValueChanged.connect(self._CORValueLocked)
        self._radioAxis.sigLockModeChanged.connect(self._modeChanged)
        self._radioAxis.sigPositionChanged.connect(self._setPositionFrmTuple)
        self._sinogramAxis.sigLockModeChanged.connect(self._modeChanged)
        self._sinogramAxis.sigReadyForComputation.connect(self._sinoAxisReady)

        # set up
        self.addPlot(self._radioAxis)
        self.addPlot(self._sinogramAxis)
        self.setActiveWidget(self._radioAxis)

    def _sinoAxisReady(self):
        self.sigSinogramReady.emit()

    def _loadCurrentSinogram(self):
        self._sinogramAxis._computationRequested()

    def _modeChanged(self):
        self.getAxis().mode = self.getMode()
        self.getAxis().use_sinogram = self.useSinogram()

    def _repeatComputationRequest(self):
        self.sigComputationRequested.emit()

    def _repeatValidateRequest(self):
        self.sigValidateRequest.emit()

    def setMode(self, mode):
        try:
            mode = AxisMode.from_value(mode)
        except:
            self.setActiveWidget(self._sinogramAxis)
            self.sigModeChanged.emit('sinogram')
        else:
            self.setActiveWidget(self._radioAxis)
            self._radioAxis.setMode(mode)
            self.sigModeChanged.emit(mode.value)

    def _CORValueLocked(self, lock):
        if lock:
            self.setMode(AxisMode.manual)
        self.setModeLock(lock)
        self.sigLockCORValueChanged.emit(lock)

    def setActiveWidget(self, widget):
        for iPlotWidget in range(self._plotWidgets.layout().count()):
            plotWidget = self._plotWidgets.layout().itemAt(iPlotWidget).widget()
            plotWidget.setVisible(widget == plotWidget)
        self.setSettingsWidget(widget.getSettingsWidget())

    def addPlot(self, plot):
        self._plotWidgets.layout().addWidget(plot)

    def setPlotWidget(self, plot):
        self.setCentralWidget(plot)

    def getPlotWidget(self):
        return self.centralWidget()

    def setSettingsWidget(self, widget):
        self._settingsDockWidget.setWidget(widget)

    def _setPositionFrmTuple(self, value):
        self.setPosition(value[0])

    def setPosition(self, value: float) -> None:
        """

        :param frm:
        :type: str
        :param value:
        :type: float

        :raises: ValueError if the frm parameter is not recognized
        """
        self._controlWidget.setPosition(value=value)

    def getAxis(self):
        return self._axis_params

    def _axisTypeChanged(self, selection):
        self._axis_params.use_sinogram = selection == 'sinogram'
        if self._axis_params.use_sinogram:
            self.setActiveWidget(self._sinogramAxis)
        else:
            self.setActiveWidget(self._radioAxis)

    def setScan(self, scan):
        """
        set the gui for this scan

        :param TomoBase scan:
        """
        self._radioAxis.setScan(scan=scan)
        self._sinogramAxis.setScan(scan=scan)
        if scan.axis_params is not None:
            self.setPosition(scan.axis_params.value)

    def _applyRequested(self) -> None:
        self.sigApply.emit()

    def _computationRequested(self) -> None:
        self.sigComputationRequested.emit()

    def hideLockButton(self) -> None:
        self._controlWidget.hideLockButton()

    def hideApplyButton(self) -> None:
        self._controlWidget.hideApplyButton()

    def setModeLock(self, lock):
        assert type(lock) is bool
        self._sinogramAxis.setLocked(locked=lock)
        self._radioAxis.setLocked(lock)

    def isModeLock(self):
        return self._radioAxis.isModeLock() or self._sinogramAxis.isModeLock()

    def isValueLock(self):
        return self._controlWidget.isValueLock()

    def _enableSinogram(self, enable):
        self._sinogramAxis.setEnabled(enable)

    def setReconsParams(self, recons_params):
        self._axis_params = recons_params
        self._radioAxis.setReconsParams(axis=recons_params)
        self._sinogramAxis.setReconsParams(axis=recons_params)

    def _axisParamsChanged(self):
        self._controlWidget.setPosition(self._axis_params.value)


class _ControlWidget(qt.QWidget):
    """
    Widget to lock cor position or compute it or validate it and to
    display the cor value
    """

    sigComputationRequest = qt.Signal()
    """Signal emitted when user request a computation from the settings"""

    sigValidateRequest = qt.Signal()
    """Signal emitted when user validate the current settings"""

    sigLockCORValueChanged = qt.Signal(bool)
    """Signal emitted when the user lock the cor value. Param: True if lock"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        # display 'center' information
        self._positionInfo = _PositionInfoWidget(parent=self)
        self.layout().addWidget(self._positionInfo)

        self._buttons = qt.QWidget(parent=self)
        self._buttons.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._buttons)

        self._lockBut = PadlockButton(parent=self)
        self._buttons.layout().addWidget(self._lockBut)
        self._lockLabel = qt.QLabel('lock cor value', parent=self)
        self._buttons.layout().addWidget(self._lockLabel)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._buttons.layout().addWidget(spacer)

        self._computeBut = qt.QPushButton('compute', parent=self)
        self._buttons.layout().addWidget(self._computeBut)
        style = qt.QApplication.style()
        applyIcon = style.standardIcon(qt.QStyle.SP_DialogApplyButton)
        self._applyBut = qt.QPushButton(applyIcon, 'validate', parent=self)
        self._buttons.layout().addWidget(self._applyBut)
        self.layout().addWidget(self._buttons)

        # set up
        self._positionInfo.setPosition(None)

        # make connection
        self._computeBut.pressed.connect(self._needComputation)
        self._applyBut.pressed.connect(self._validate)
        self._lockBut.sigLockChanged.connect(self._lockValueChanged)

    def hideLockButton(self) -> None:
        self._lockLabel.hide()
        self._lockBut.hide()

    def hideApplyButton(self) -> None:
        self._applyBut.hide()

    def _lockValueChanged(self):
        self.sigLockCORValueChanged.emit(self._lockBut.isLocked())
        self._computeBut.setEnabled(not self._lockBut.isLocked())

    def _needComputation(self, *arg, **kwargs):
        """callback when the radio line changed"""
        self.sigComputationRequest.emit()

    def _validate(self):
        self.sigValidateRequest.emit()

    def setPosition(self, value):
        self._positionInfo.setPosition(value=value)

    def isValueLock(self):
        return self._lockBut.isLocked()


class _PositionInfoWidget(qt.QWidget):
    """Widget used to display information relative to the current position"""

    def __init__(self, parent, axis=None):
        self._axis = None
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        centerLabel = qt.QLabel('center:', parent=self)
        centerLabel.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        font = centerLabel.font()
        font.setBold(True)
        centerLabel.setFont(font)

        self.layout().addWidget(centerLabel)
        self._positionLabel = qt.QLabel('', parent=self)
        self._positionLabel.setSizePolicy(qt.QSizePolicy.Expanding,
                                          qt.QSizePolicy.Minimum)
        palette = self._positionLabel.palette()
        palette.setColor(qt.QPalette.WindowText, qt.QColor(qt.Qt.red))
        self._positionLabel.setPalette(palette)
        self.layout().addWidget(self._positionLabel)

        if axis:
            self.setAxis(axis)

    def setAxis(self, axis):
        assert isinstance(axis, QAxisRP)
        if axis == self._axis:
            return
        if self._axis is not None:
            self._axis.sigChanged.disconnect(self._updatePosition)
        self._axis = axis
        self._axis.sigChanged.connect(self._updatePosition)
        self._updatePosition()

    def _updatePosition(self):
        if self._axis:
            if self._axis.value is None:
                value = '?'
            else:
                value = str(self._axis.value)
            self._positionLabel.setText(value)

    def getPosition(self):
        return float(self._positionLabel.text())

    def setPosition(self, value):
        if value is None:
            self._positionLabel.setText('?')
        else:
            self._positionLabel.setText(str(value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import argparse
from silx.gui import qt
import signal
import functools
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.reconstruction.axis import AxisWindow
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.synctools.axis import QAxisRP
from tomwer.core.process.reconstruction.axis.axis import AxisProcess, NoAxisUrl
from silx.io.url import DataUrl
import logging


logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


class _AxisProcessGUI(AxisProcess, AxisWindow):
    def __init__(self, scan, axis_params):
        AxisProcess.__init__(self, axis_params=axis_params)
        AxisWindow.__init__(self, axis_params=axis_params)
        self.hideLockButton()
        self.hideApplyButton()
        self.setScan(scan=scan)

        # connect Signal / Slot
        callback = functools.partial(self.process, scan)
        self.sigComputationRequested.connect(callback)
        self.setWindowIcon(icons.getQIcon('tomwer'))

    def compute(self, scan, wait=True):
        mess = ' '.join(('start axis calculation with',
                         scan.axis_params.axis_url_1.url.path()))
        _logger.info(mess)
        try:
            AxisProcess.compute(self, scan=scan, wait=wait)
        except NoAxisUrl:
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = "Unable to find url to compute the axis, please select them " \
                   "from the `axis input` tab"
            msg.setText(text)
            msg.exec_()
            return None
        else:
            position = scan.axis_params.value
            assert isinstance(position, (float, type(None)))
            frm = 'sinogram' if scan.axis_params.use_sinogram else 'radio'
            AxisWindow.setPosition(self, frm=frm, value=position)
            return position

    def _updatePosition(self, scan):
        self._widget.setPosition(frm=None, value=scan.axis_params.value)


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--scan-path',
        help='For EDF acquisition: provide folder path, for HDF5 / nexus'
             'provide the master file',
        default=None)
    parser.add_argument(
        '--use-sinogram',
        help='use the signoram from radio for computing COR. This only work'
             'with the =scan-path option',
        action="store_true",
        default=False)
    parser.add_argument(
        '--radio1',
        help='If you cannot provide scan-path or master file you should ' \
             'provide urls of the two radios to use for calculation',
        default=None)
    parser.add_argument(
        '--radio2',
        help='If you cannot provide scan-path or master file you should ' \
             'provide urls of the two radios to use for calculation',
        default=None)
    parser.add_argument(
        '--entry',
        help='For Nexus files: entry in the master file',
        default=None,
    )
    parser.add_argument(
        '--debug',
        dest="debug",
        action="store_true",
        default=False,
        help='Set logging system in debug mode')

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    if options.scan_path is not None:
        if os.path.isdir(options.scan_path):
            options.scan_path = options.scan_path.rstrip(os.path.sep)
            scan = ScanFactory.create_scan_object(scan_path=options.scan_path)
        else:
            if not os.path.isfile(options.scan_path):
                raise ValueError('scan path should be a folder containing an'
                                 ' EDF acquisition or an hdf5 - nexus '
                                 'compliant file')
            if options.entry is None:
                raise ValueError('entry in the master file should be specify')
            scan = HDF5TomoScan(scan=options.scan_path,
                                entry=options.entry)
    else:
        scan = ScanFactory.mock_scan()
        if options.radio1 is None:
            raise ValueError('At least a file or a scan path should be given')
    if scan.axis_params is None:
        scan.axis_params = QAxisRP()
    if options.radio1 is not None:
        radio1 = DataUrl(file_path=options.radio1, scheme='fabio')
        scan.axis_params.axis_url_1 = radio1
        if options.radio2 is not None:
            radio2 = DataUrl(file_path=options.radio2, scheme='fabio')
        else:
            radio2 = None
        scan.axis_params.axis_url_2 = radio2

    if options.debug:
        _logger.setLevel(logging.DEBUG)

    axis_params = QAxisRP()
    axis_params.use_sinogram = options.use_sinogram
    if scan.dim_2 is not None:
        axis_params.sinogram_line = scan.dim_2 // 2

    window = _AxisProcessGUI(scan=scan, axis_params=axis_params)
    window.setWindowTitle('axis')
    window.setWindowIcon(icons.getQIcon('tomwer'))

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.exec_()


def getinputinfo():
    return "tomwer axis [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == '__main__':
    main(sys.argv)



#/*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/02/2017"

"""Module defining the dition of a single H5Struct"""

from tomwer.core.log import TomwerLogger
from tomwer.gui.reconstruction.ftserie.h5editor import H5StructsEditor

logger = TomwerLogger(__name__)


class H5StructEditor(H5StructsEditor):
    """simple class which link an interface with an h5 struct used for pyhst"""

    def __init__(self, structID):
        self.structID = structID
        H5StructsEditor.__init__(self, (self.structID,))

    def linkComboboxWithH5Variable(self, qcombobox, h5ParamName,
                                   fitwithindex=False, dic=None, filters=None,
                                   setDefault=True):
        """

        :param qcombobox: the QCombobox in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param fitwithindex: true if we want to store the index in the h5File,
            otherwise store the string
        :param filters: filter to apply on the string to get the requested
            value
        """
        return H5StructsEditor.linkComboboxWithH5Variable(self,
                                                          qcombobox=qcombobox,
                                                          structID=self.structID,
                                                          h5ParamName=h5ParamName,
                                                          fitwithindex=fitwithindex,
                                                          dic=dic,
                                                          filters=filters,
                                                          setDefault=setDefault)

    def linkCheckboxWithH5Variable(self, qcheckbox, h5ParamName, invert=False):
        """

        :param qcheckbox: the QCheckBox in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param invert: true if we will not store the value state == checked but
            state == unchecked
        :param structID: the structID associated with the param ID
        """
        return H5StructsEditor.linkCheckboxWithH5Variable(self,
                                                          qcheckbox=qcheckbox,
                                                          structID=self.structID,
                                                          h5ParamName=h5ParamName,
                                                          invert=invert)

    def LinkLineEditWithH5Variable(self, qlineedit, h5ParamName,
                                   h5paramtype=str):
        """

        :param qlineedit: the QLineEdit in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param h5ParamName: the parameter ID in the H5 structure.
        """
        return H5StructsEditor.LinkLineEditWithH5Variable(self,
                                                          qlineedit=qlineedit,
                                                          h5ParamName=h5ParamName,
                                                          structID=self.structID,
                                                          h5paramtype=h5paramtype)

    def LinkSelectionLineEditWithH5Variable(self, qlineedit, h5ParamName,
                                            h5paramtype=str):
        """

        :param qlineedit: the QLineEdit in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param h5ParamName: the parameter ID in the H5 structure.
        """
        return H5StructsEditor.LinkSelectionLineEditWithH5Variable(self,
                                                                   qlineedit=qlineedit,
                                                                   h5ParamName=h5ParamName,
                                                                   structID=self.structID,
                                                                   h5paramtype=h5paramtype)

    def linkGroupWithH5Variable(self, group, h5ParamName, getter, setter):
        return H5StructsEditor.linkGroupWithH5Variable(self,
                                                       group=group,
                                                       structID=self.structID,
                                                       h5ParamName=h5ParamName,
                                                       getter=getter,
                                                       setter=setter)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno", "T. Vincent"]
__license__ = "MIT"
__date__ = "04/01/2018"


try:
    from silx.gui import qt
except ImportError:
    raise ImportError("Can't found silx modules")
from tomwer.gui import icons
import tomwer.version


def getMainSplashScreen():
    pixmap = icons.getQPixmap('tomwer')
    splash = qt.QSplashScreen(pixmap)
    splash.show()
    splash.raise_()
    _version = tomwer.version.version
    text = 'version ' + str(_version)
    splash.showMessage(text, qt.Qt.AlignLeft | qt.Qt.AlignBottom, qt.Qt.white)
    return splash



# coding: utf-8
#/*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"


from silx.gui import qt

from tomwer.core.process.conditions import filters
from tomwer.gui import icons
from tomwer.gui.utils.sandboxes import RegularExpressionSandBox, \
    RegularExpressionSandBoxDialog


class FileNameFilterWidget(filters.FileNameFilter, qt.QWidget):
    """
    Simple widget allowing the user to define a pattern and emitting a signal
    'sigValid' or 'sigUnvalid' if passing through the filter or not
    """
    sigValid = qt.Signal(str)
    """Signal emitted when the string pass through the filter"""
    sigUnvalid = qt.Signal(str)
    """Signal emitted when the string doesn't pass through the filter"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        filters.FileNameFilter.__init__(self, None)

        self.setLayout(qt.QGridLayout())

        self._invertCB = qt.QCheckBox('Invert filter action', parent=self)
        self._invertCB.setToolTip('If not inverted and match the filter '
                                  'condition then let the scan pass through')
        self.layout().addWidget(self._invertCB, 0, 0, 1, 3)

        self.layout().addWidget(qt.QLabel('filter type:'), 1, 0)
        self._filterTypeCB = qt.QComboBox(parent=self)
        for filter_type in self.FILTER_TYPES:
            self._filterTypeCB.addItem(filter_type)
        self.layout().addWidget(self._filterTypeCB, 1, 1, 1, 2)

        self.layout().addWidget(qt.QLabel('pattern:', parent=self), 2, 0)
        self._patternLE = qt.QLineEdit(parent=self)
        self._patternLE.setToolTip(self.getPatternTooltip())
        self.layout().addWidget(self._patternLE, 2, 1)

        icon = icons.getQIcon('information')
        self._sandBoxPB = qt.QPushButton(icon=icon, parent=self)
        # self._sandBoxPB.setPixmap(icon.pixmap(qt.QSize(32, 32)))
        self._sandBoxPB.setToolTip(RegularExpressionSandBox.description())
        self._sandBoxPB.pressed.connect(self._showSandBox)
        self.layout().addWidget(self._sandBoxPB, 2, 2)

        self._sandBoxPB.setVisible(False)

        self.sandboxDialog = None

        # signal/SLOT connection
        if qt.BINDING in ('PyQt4', 'PySide'):
            self._filterTypeCB.currentIndexChanged[str].connect(self.setActiveFilter)
        else:
            self._filterTypeCB.currentTextChanged[str].connect(self.setActiveFilter)

    def isFiltered(self, value):
        """
        Check if the name go thought the filter. Will emit either 'sigvalid' or
        'sigUnvalid' signals according to the result

        :param str value: value to check
        :return: True if pass by the filter
        """
        assert type(value) is str
        _filtered = super().isFiltered(value)
        if self._invertCB.isChecked():
            _filtered = not _filtered
        if _filtered is True:
            self.sigUnvalid.emit(value)
        else:
            self.sigValid.emit(value)
        return _filtered

    def _showSandBox(self):
        self.getSandboxDialog().show()

    def getSandboxDialog(self):
        if self.sandboxDialog is None:
            self.sandboxDialog = RegularExpressionSandBoxDialog(parent=None,
                                                                pattern=self.getPattern())
            self._patternLE.editingFinished.connect(self._updateSandBoxPattern)
            self.sandboxDialog.setModal(False)
        return self.sandboxDialog

    def getPatternTooltip(self):
        return 'define the pattern to accept file using the python `re` ' \
               'library.\n' \
               'For example if we two series of acquisition named serie_10_XXX\n' \
               'and serie_100_YYY and we want to filter all the serie_10_XXX\n' \
               'then we can define the pattern "serie_100_" \n' \
               'which will only let the serie_100_YYY go through.'

    def _updateSandBoxPattern(self):
        self.getSandboxDialog().setPattern(self._patternLE.text())

    def unvalidPatternDefinition(self, pattern, error):
        """Overwrite NameFilter.unvalidPatternDefinition"""
        title = 'regular expression %s is invalid' % pattern
        mess = qt.QMessageBox(qt.QMessageBox.warning, title, parent=self,
                              text=str(error))
        mess.setModal(False)
        mess.show()

    def setActiveFilter(self, filter):
        self.activeFilter = filter
        self._sandBoxPB.setVisible(filter == 'regular expression')



# coding: utf-8
#/*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to manage observations. Initially on files.
Observations are runned on a thread and run each n seconds.
They are manage by thread and signals
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "09/02/2017"

import threading
import time


class WaiterThread(threading.Thread):
    """simple thread wich wait for waitingTime to be finished"""

    finishedEvent = threading.Event()
    """Event registred when the thread is finished"""

    def __init__(self, waitingTime):
        super(WaiterThread, self).__init__()
        self.waitingTime = waitingTime

    def run(self):
        time.sleep(self.waitingTime)
        self.finishedEvent.set()
